from typing import *
from enum import Enum, auto

from base_aux.base_statics.m2_exceptions import *


# =====================================================================================================================
TYPE__BREED_RESULT__ITEM = Union[Any, Exception]
TYPE__BREED_RESULT__GROUP = Union[
    TYPE__BREED_RESULT__ITEM,        # SINGLE variant
    list[TYPE__BREED_RESULT__ITEM]   # LIST variant
]
TYPE__BREED_RESULT__GROUPS = dict[str, TYPE__BREED_RESULT__GROUP]


# =====================================================================================================================
class BreederObjectList_GroupType(Enum):
    SINGLE = auto()
    LIST = auto()
    NOT_EXISTS = auto()


# =====================================================================================================================
class BreederObjectList:
    """
    class which keep all aux_types in one place!
    useful for multyObject systems.

    If you need just one object_instance for all duts - use direct attribute,
    else use LIST__*NAME* and dont forget to create annotation for direct Indexed item access!

    so you could
    - pass just one instance into all other classes!
    - check all devices for PRESENT (or else) in one place!
    - init all and check correctness for all

    AFTER GENERATING OBJECTS - ACCESS TO OBJECTS LIST USED OVER THE CLASS!!!
        OBJS_CLS = BreederObjectList
        OBJS = OBJS_CLS()
        devs = OBJS_CLS.LIST__DEV
    """
    # SETTINGS ----------------------
    COUNT: int = 1

    # usage EXAMPLES ------------------------------------------------------
    # CLS_LIST__DUT: type[DutBase] = DutBase
    # LIST__DUT: list[DutBase]
    # DUT: DutBase

    # CLS_SINGLE__ATC: Callable[..., DeviceBase]
    # ATC: DeviceBase

    # AUX ----------------------------------------------------------
    # definitions -----
    _STARTSWITH__DEFINE__CLS_LIST: str = "CLS_LIST__"
    _STARTSWITH__DEFINE__CLS_SINGLE: str = "CLS_SINGLE__"

    # access ----------
    _STARTSWITH__ACCESS__OBJECT_LIST__IN_BREEDER: str = "LIST__"
    _STARTSWITH__ACCESS__OBJECT_LIST__IN_SOURCE: str = "INSTS"
    _STARTSWITH__ACCESS__BREEDER_IN__SOURCE: str = "BREEDER"    # SINGLE:type[Breeder] // LIST:Breeder

    # AUX --------
    __groups__are_generated: bool = False

    # instance ---
    INDEX: int | None = None    # index used only in OBJECT INSTANCE

    def __init__(self, index: int, *args, **kwargs):
        """
        init only when you need to do access to exact items!
        """
        self.INDEX = index      # need first!
        super().__init__(*args, **kwargs)
        # self.generate__objects()

    @classmethod
    def groups_check__generated(cls) -> bool | None:
        return cls.__groups__are_generated

    # -----------------------------------------------------------------------------------------------------------------
    @classmethod
    def generate__objects(cls, force: bool | None = None) -> None:
        """exact and only one method to Gen all aux_types - dont forget to call it!
        """
        if force:
            cls.__groups__are_generated = False

        if cls.__groups__are_generated:
            return

        # WORK --------------------------------------
        for attr_name in dir(cls):
            # LIST --------------------------------------
            if attr_name.startswith(cls._STARTSWITH__DEFINE__CLS_LIST):
                group_name = attr_name.removeprefix(cls._STARTSWITH__DEFINE__CLS_LIST)
                obj_cls = getattr(cls, attr_name)

                obj_list__name = f"{cls._STARTSWITH__ACCESS__OBJECT_LIST__IN_BREEDER}{group_name}"
                obj_list__value = []
                for index in range(cls.COUNT):
                    try:
                        obj_instance = obj_cls(index)
                        setattr(obj_instance, cls._STARTSWITH__ACCESS__BREEDER_IN__SOURCE, cls(index=index))
                    except Exception as exx:
                        obj_instance = exx
                    obj_list__value.append(obj_instance)


                # apply GROUP to class -------
                setattr(cls, obj_list__name, obj_list__value)
                setattr(obj_cls, cls._STARTSWITH__ACCESS__OBJECT_LIST__IN_SOURCE, obj_list__value)

            # SINGLE --------------------------------------
            if attr_name.startswith(cls._STARTSWITH__DEFINE__CLS_SINGLE):
                group_name = attr_name.removeprefix(cls._STARTSWITH__DEFINE__CLS_SINGLE)
                obj_cls = getattr(cls, attr_name)
                try:
                    obj_instance = obj_cls()
                except Exception as exx:
                    obj_instance = exx
                # apply -------
                setattr(cls, group_name, obj_instance)
                setattr(obj_cls, cls._STARTSWITH__ACCESS__OBJECT_LIST__IN_SOURCE, obj_instance)
                setattr(obj_cls, cls._STARTSWITH__ACCESS__BREEDER_IN__SOURCE, cls)

        cls.__groups__are_generated = True

    # -----------------------------------------------------------------------------------------------------------------
    @classmethod
    def groups__get_names(cls) -> set[str]:
        result = set()
        for attr_name in dir(cls):
            if attr_name.startswith(cls._STARTSWITH__DEFINE__CLS_LIST):
                group_name = attr_name.removeprefix(cls._STARTSWITH__DEFINE__CLS_LIST)
                result.update([group_name, ])
            if attr_name.startswith(cls._STARTSWITH__DEFINE__CLS_SINGLE):
                group_name = attr_name.removeprefix(cls._STARTSWITH__DEFINE__CLS_SINGLE)
                result.update([group_name, ])

        return result

    # -----------------------------------------------------------------------------------------------------------------
    @classmethod
    def groups_count__generated(cls) -> int | None:
        """
        work only after called generate__objects(),
        so if you wasnot call generate__objects it will return None!
        """
        if cls.__groups__are_generated:
            return len(cls.groups__get_names())

    # GROUP -----------------------------------------------------------------------------------------------------------
    @classmethod
    def group_get__type(cls, name: str) -> BreederObjectList_GroupType:
        if f"{cls._STARTSWITH__DEFINE__CLS_SINGLE}{name}" in dir(cls):
            return BreederObjectList_GroupType.SINGLE

        if f"{cls._STARTSWITH__DEFINE__CLS_LIST}{name}" in dir(cls):
            return BreederObjectList_GroupType.LIST

        return BreederObjectList_GroupType.NOT_EXISTS

    @classmethod
    def group_check__exists(cls, name: str) -> bool:
        return cls.group_get__type(name) != BreederObjectList_GroupType.NOT_EXISTS

    @classmethod
    def group_get__cls(cls, name: str) -> Any | None:
        group_type = cls.group_get__type(name)

        if group_type == BreederObjectList_GroupType.SINGLE:
            attr = f"{cls._STARTSWITH__DEFINE__CLS_SINGLE}{name}"
            return getattr(cls, attr)

        if group_type == BreederObjectList_GroupType.LIST:
            attr = f"{cls._STARTSWITH__DEFINE__CLS_LIST}{name}"
            return getattr(cls, attr)

    @classmethod
    def group_get__insts(cls, name: str) -> Union[None, Any, list[Any]]:
        if cls.group_check__exists(name) and cls.__groups__are_generated:
            group_cls = cls.group_get__cls(name)
            result = getattr(group_cls, cls._STARTSWITH__ACCESS__OBJECT_LIST__IN_SOURCE)
            return result

    @classmethod
    def group_call__(cls, meth: str, group: str | None = None, args: list | None = None, kwargs: dict | None = None) -> Union[NoReturn, TYPE__BREED_RESULT__GROUP, TYPE__BREED_RESULT__GROUPS]:
        """
        call one method on exact group (every object in group) or all groups (every object in all groups).
        created specially for call connect/disconnect for devices in TP.

        :param meth:
        :param group:

        :param args:
        :param kwargs:
        :return:
            RAISE only if passed group and group is not exists! or groups are not generated
        """
        if not cls.__groups__are_generated:
            raise Exx__BreederObjectList_GroupsNotGenerated()

        args = args or ()
        kwargs = kwargs or {}

        # CALL ON ALL GROUPS -------------------------------------------------
        if group is None:
            results = {}
            for group_name in cls.groups__get_names():
                results.update({group_name: cls.group_call__(meth, group_name, args, kwargs)})
            return results

        # if group is not exists ---------------------------------------------
        if not cls.group_check__exists(group):
            raise Exx__BreederObjectList_GroupNotExists(group)

        # EXACT ONE GROUP ----------------------------------------------------
        group_objs = cls.group_get__insts(group)

        if isinstance(group_objs, list):
            results = []
            for obj in group_objs:
                try:
                    obj_meth = getattr(obj, meth)
                    obj_result = obj_meth(*args, **kwargs)
                except Exception as exx:
                    obj_result = exx
                results.append(obj_result)
        else:
            obj = group_objs
            try:
                obj_meth = getattr(obj, meth)
                obj_result = obj_meth(*args, **kwargs)
            except Exception as exx:
                obj_result = exx
            results = obj_result

        return results

    # -----------------------------------------------------------------------------------------------------------------
    def __getattr__(self, item: str) -> Union[None, Any, NoReturn]:
        """
        :param item:
        :return: existed OBJECT from LIST or SINGLE!!!
        """
        if self.INDEX is None:
            return

        # ACCESS TO OBJECT ----------------------------
        group_insts = self.group_get__insts(item)
        if group_insts is not None:
            if isinstance(group_insts, list):
                inst = group_insts[self.INDEX]
            else:
                inst = group_insts
            return inst

        # FINAL not found -----------------------------
        msg = f"{item=}/{self.INDEX=}"
        print(msg)
        raise Exx__BreederObjectList_GroupNotExists(msg)


# =====================================================================================================================



from typing import *

from base_aux.lambdas.m1_lambdas import Lambda
from base_aux.aux_attr.m1_attr2_nest2_lambdas_resolve import NestInit_AttrsLambdaResolve
from base_aux.privates.m1_privates import *
from base_aux.aux_attr.m4_kits import *
from .m0_base import *


# =====================================================================================================================
class RecipientTgID(Nest_AttrKit):
    MyTgID: str


# =====================================================================================================================
class AlertTelegram(NestInit_AttrsLambdaResolve, AlertBase):
    """realisation for sending Telegram msg
    """
    # SETTINGS ------------------------------------
    SERVER_TG: AttrKit_AuthNamePwd = PvLoaderIni_AuthTgBot(keypath=("TGBOT_DEF",))
    RecipientTgID: AttrKit_AuthNamePwd = PvLoaderIni(target=RecipientTgID, keypath=("TG_ID",))

    # AUX -----------------------------------------
    _conn: telebot.TeleBot

    def _connect_unsafe(self) -> Union[bool, NoReturn]:
        self._conn = telebot.TeleBot(token=self.SERVER_TG.TOKEN)
        return True

    def _send_unsafe(self) -> Union[bool, NoReturn]:
        self._conn.send_message(chat_id=self.RECIPIENT, text=self._msg_compose())
        return True

    def _msg_compose(self) -> str:
        msg = f"{self.SUBJECT}\n{self.body}"
        return msg

    def _recipient_self_get(self) -> str:
        return self.RecipientTgID.MyTgID


# =====================================================================================================================



# =====================================================================================================================
# USE COMMON/GENERAL TYPES

_std = [
    # base ----------------
    BaseException,
    Exception,
    BaseExceptionGroup,

    # imports -------------
    ImportError,
    ImportWarning,
    ModuleNotFoundError,

    # FILE/PATH
    FileExistsError,    # ExistsAlready
    FileNotFoundError,  # NotExists

    NotADirectoryError,
    IsADirectoryError,

    # USER ----------------
    UserWarning,
    Warning,
    DeprecationWarning,
    PendingDeprecationWarning,

    AssertionError,

    NotImplemented,
    NotImplementedError,

    # VALUE ---------------
    # type
    TypeError,

    # value
    ValueError,

    # syntax/format
    SyntaxWarning,
    SyntaxError,
    IndentationError,

    EOFError,
    TabError,
    BytesWarning,

    EncodingWarning,

    UnicodeWarning,
    UnicodeDecodeError,
    UnicodeEncodeError,
    UnicodeTranslateError,

    # ACCESS ------
    NameError,
    AttributeError,
    PermissionError,
    KeyError,
    IndexError,

    # COLLECTION
    GeneratorExit,
    StopIteration,
    StopAsyncIteration,

    # arithm/logic
    ZeroDivisionError,
    ArithmeticError,
    FloatingPointError,
    OverflowError,

    RecursionError,
    BrokenPipeError,
    InterruptedError,

    # CONNECTION
    ConnectionError,
    ConnectionAbortedError,
    ConnectionResetError,
    ConnectionRefusedError,
    TimeoutError,

    # OS/OTHER
    SystemExit,
    WindowsError,
    IOError,
    OSError,
    EnvironmentError,
    SystemError,
    ChildProcessError,
    MemoryError,
    KeyboardInterrupt,

    BufferError,
    LookupError,

    UnboundLocalError,

    RuntimeWarning,
    ResourceWarning,
    ReferenceError,
    ProcessLookupError,
    RuntimeError,
    FutureWarning,
    ExceptionGroup,
    BlockingIOError,

    # REAL VALUE = NOT AN EXCEPTION!!!
    NotImplemented,      # NotImplemented = None # (!) real value is 'NotImplemented'
]


# =====================================================================================================================
class Exx__BoolBase(Exception):
    """
    just a solution to get correct bool() if get Exx as value

    SPECIALLY CREATED FOR
    ---------------------
    classes.VALID if
    """
    def __bool__(self):
        return False


# =====================================================================================================================
class Exx__WrongUsage(Exx__BoolBase):
    """
    GOAL
    ----
    you perform incorrect usage!

    SPECIALLY CREATED FOR
    ---------------------
    NoValue - never instantiate it! use value only as Class!
    """


class Exx__Incompatible(Exx__BoolBase):
    pass


class Exx__OutOfRange(Exx__BoolBase):
    pass


# =====================================================================================================================
class Exx__AnnotNotDefined(Exx__BoolBase):
    """Exception in case of not defined attribute in instance
    """


class Exx__NumberArithm_NoName(Exx__BoolBase):
    pass


class Exx__GetattrPrefix(Exx__BoolBase):
    pass


class Exx__GetattrPrefix_RaiseIf(Exx__GetattrPrefix):
    pass


class Exx__ValueNotParsed(Exx__BoolBase):
    pass


class Exx__ValueUnitsIncompatible(Exx__BoolBase):
    pass


class Exx__IndexOverlayed(Exx__BoolBase):
    pass


class Exx__IndexNotSet(Exx__BoolBase):
    pass


class Exx__NotExists(Exx__BoolBase):
    pass


class Exx__StartOuterNONE_UsedInStackByRecreation(Exx__BoolBase):
    """
    in stack it will be recreate automatically! so dont use in pure single BreederStrSeries!
    """
    pass


class Exx__BreederObjectList_GroupsNotGenerated(Exx__BoolBase):
    pass


class Exx__BreederObjectList_GroupNotExists(Exx__BoolBase):
    pass


class Exx__BreederObjectList_ObjCantAccessIndex(Exx__BoolBase):
    pass


# =====================================================================================================================
class Exx__Valid(Exx__BoolBase):
    pass


class Exx__ValueNotValidated(Exx__Valid):
    pass


# =====================================================================================================================
class Exx__SameKeys(Exx__BoolBase):
    """Same keys NOT allowed!
    """


# =====================================================================================================================
if __name__ == '__main__':
    # REASON --------------
    assert bool(Exception(0)) is True
    assert bool(Exception(False)) is True

    # SOLUTION --------------
    assert bool(Exx__BoolBase(0)) is False
    assert bool(Exx__BoolBase(False)) is False


# =====================================================================================================================



from typing import *

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from base_aux.pyqt.m1_dialog import DialogsSet


# SET ============================================================================================================
class DialogsSetTp(DialogsSet):
    """
    attempt to keep all available dialogs for current project in one place!
    so to be sure there are no any other available!
    """
    @staticmethod
    def info__about(*args) -> int:
        wgt = QMessageBox()
        wgt.setMaximumWidth(1000)
        answer = wgt.information(
            None,
            "О программе",
            (
                "ООО Элемент-Инжиниринг,\n"
                "Программа проведения тестирования блоков питания\n"
                "(стендовые испытания ОТК)"
             )
        )
        # return always 1024
        return answer

    @staticmethod
    def finished__devs_detection(*args) -> int:
        wgt = QMessageBox()
        wgt.resize(1000, 1000)
        wgt.setBaseSize(1000, 1000)
        wgt.setToolTip("hello")
        answer = wgt.information(
            None,
            "Определение устройств",
            (
                "Процесс завершен" + " "*30
            )
        )
        # return always 1024
        return answer

    @staticmethod
    def finished__tp(*args) -> int:
        answer = QMessageBox.information(
            None,
            "Тестирование",
            (
                "Процесс завершен"
             )
        )
        # return always 1024
        return answer

    @staticmethod
    def finished__save(*args) -> int:
        answer = QMessageBox.information(
            None,
            "Сохранение",
            (
                "Процесс завершен"
             )
        )
        # return always 1024
        return answer


# =====================================================================================================================
if __name__ == '__main__':
    # DialogsSetTp.info__about()
    DialogsSetTp.finished__devs_detection()
    # DialogsSetTp.finished__tp()
    # DialogsSetTp.finished__save()


# =====================================================================================================================



import sys
from typing import Union, Any, Callable, NoReturn

from base_aux.aux_attr.m3_getattr1_prefix_1_inst import GetattrPrefixInst_RaiseIf
from base_aux.aux_types.m1_type_aux import TypeAux
from base_aux.versions.m2_version import Version


# =====================================================================================================================
class CheckVersion(GetattrPrefixInst_RaiseIf):
    SOURCE: Union[Any, Callable[..., Any]] = sys.version.split()[0]
    # print(sys.version_info)   # sys.version_info(major=3, minor=8, micro=10, releaselevel='final', serial=0)
    # return sys.version_info[:3]     # (3, 8, 10)

    # -----------------------------------------------------------------------------------------------------------------
    def __init__(self, source: Any | Callable | None = None):
        if source is not None:
            self.SOURCE = source

    @property
    def ACTUAL(self) -> Version:
        if TypeAux(self.SOURCE).check__callable_func_meth_inst():
            value = self.SOURCE()
        else:
            value = self.SOURCE

        return Version(value)

    # ---------------------------------------
    def check_eq(self, target: Any):
        return self.ACTUAL == target

    def check_ne(self, target: Any):
        return self.ACTUAL != target

    # ---------------------------------------
    def check_le(self, target: Any):
        return self.ACTUAL <= target

    def check_lt(self, target: Any):
        return self.ACTUAL < target

    # ---------------------------------------
    def check_ge(self, target: Any):
        return self.ACTUAL >= target

    def check_gt(self, target: Any):
        return self.ACTUAL > target


# =====================================================================================================================
class CheckVersion_Python(CheckVersion):
    """
    check version of python interpreter.

    USAGE
    -----
    CheckVersion_Python().raise_if_not__check_ge("2")
    CheckVersion_Python().raise_if_not__check_ge("3.11")
    CheckVersion_Python().raise_if_not__check_ge("3.11rc1", _comment="need Python GRATER EQUAL")
    """
    SOURCE = sys.version.split()[0]

    raise_if__check_eq: Callable[..., NoReturn | None]
    raise_if_not__check_eq: Callable[..., NoReturn | None]

    raise_if__check_ne: Callable[..., NoReturn | None]
    raise_if_not__check_ne: Callable[..., NoReturn | None]

    raise_if__check_le: Callable[..., NoReturn | None]
    raise_if_not__check_le: Callable[..., NoReturn | None]

    raise_if__check_lt: Callable[..., NoReturn | None]
    raise_if_not__check_lt: Callable[..., NoReturn | None]

    raise_if__check_ge: Callable[..., NoReturn | None]
    raise_if_not__check_ge: Callable[..., NoReturn | None]

    raise_if__check_gt: Callable[..., NoReturn | None]
    raise_if_not__check_gt: Callable[..., NoReturn | None]


# =====================================================================================================================



from base_aux.base_statics.m2_exceptions import *
from base_aux.aux_types.m1_type_aux import *
from base_aux.base_inits.m1_nest_init_source import *
from base_aux.base_statics.m1_types import *

from base_aux.aux_argskwargs.m1_argskwargs import *
from base_aux.aux_iter.m1_iter_aux import *
from base_aux.aux_attr.m1_attr2_nest1_gsai_anycase import *


# =====================================================================================================================
@final
class AnnotsAux(NestInit_Source):
    """
    GOAL
    ----
    work with all __annotations__
        from all nested classes
        in correct order

    RULES
    -----
    4. nesting available with correct order!
        class ClsFirst(BreederStrStack):
            atr1: int
            atr3: int = None

        class ClsLast(BreederStrStack):
            atr2: int = None
            atr4: int

        for key, value in ClsLast.annotations__get_nested().items():
            print(f"{key}:{value}")

        # atr1:<class 'int'>
        # atr3:<class 'int'>
        # atr2:<class 'int'>
        # atr4:<class 'int'>
    """
    # -----------------------------------------------------------------------------------------------------------------
    def get_not_defined(self) -> list[str]:
        """
        GOAL
        ----
        return list of not defined annotations

        SPECIALLY CREATED FOR
        ---------------------
        annot__check_all_defined
        """
        result = []
        nested = self.get__dict_types()
        for key in nested:
            if not AttrAux(self.SOURCE).anycase__check_exists(key):
                result.append(key)
        return result

    # =================================================================================================================
    def check_all_defined(self) -> bool:
        """
        GOAL
        ----
        check if all annotated aux_attr have value!
        """
        return not self.get_not_defined()

    def check_all_defined_or_raise(self) -> None | NoReturn:
        """
        GOAL
        ----
        check if all annotated aux_attr have value!
        """
        not_defined = self.get_not_defined()
        if not_defined:
            dict_type = self.get__dict_types()
            msg = f"[CRITICAL]{not_defined=} in {dict_type=}"
            raise Exx__AnnotNotDefined(msg)

    # =================================================================================================================
    def set_value(self, name: str, value: Any = None, only_annot: bool = True) -> None:   # FIXME: decide -  DONT DELETE only_annot!!! it could be used like filter?
        """
        GOAL
        ----
        set name/value by intended name from Annots or direct Attrs! or even not existsed!
        without creation new annot!
        """
        if self.name__check_exists(name):
            name = self.name__get_original(name)    # replace by original
        elif only_annot:
            return

        AttrAux(self.SOURCE).anycase__setattr(name, value)

    def set_values__by_dict(self, data: TYPING.KWARGS_FINAL, only_annot: bool = True) -> None:
        """
        GOAL
        ----
        SAME AS AttrAux.SetValues but assume names from annots!
        # set attrs for annotated names (fixme: ???? AND STD ATTRS! - it seems OK!
        # so annots+attrs used like filter

        SPECIALLY CREATED FOR
        ---------------------
        load into AnnotTemplate
        """
        for key, value in data.items():
            self.set_value(key, value, only_annot=only_annot)

    # -----------------------------------------------------------------------------------------------------------------
    def set_annots_attrs__by_args_kwargs(self, *args: Any, **kwargs: TYPING.KWARGS_FINAL) -> None:
        """
        CREATED SPECIALLY FOR
        ---------------------
        NestInit_AnnotsAttrByKwArgs
        """
        self.set_annots__by_args(*args)
        self.set_annots_attrs__by_kwargs(**kwargs)

    def set_annots_attrs__by_kwargs(self, **kwargs: TYPING.KWARGS_FINAL) -> None:
        return self.set_values__by_dict(kwargs, only_annot=False)

    # -----------------------------------------------------------------------------------------------------------------
    def set_annots__by_args_kwargs(self, *args: Any, **kwargs: TYPING.KWARGS_FINAL) -> None:
        """
        CREATED SPECIALLY FOR
        ---------------------
        NestInit_AnnotsAttrByKwArgs
        """
        self.set_annots__by_args(*args)
        self.set_annots__by_kwargs(**kwargs)

    def set_annots__by_kwargs(self, **kwargs: TYPING.KWARGS_FINAL) -> None:
        return self.set_values__by_dict(kwargs, only_annot=True)

    def set_annots__by_args(self, *args: Any) -> None:
        """
        ARGS - ARE VALUES! not names!

        IF ARGS MORE then Annots - NoRaise! You should lnow what you do!    # FIXME: decide?
        """
        return self.set_annots_attrs__by_kwargs(**dict(zip(self.iter_names(), args)))

    # =================================================================================================================
    def get__dict_types(self) -> dict[str, type[Any]]:
        """
        GOAL
        ----
        get all annotations in correct order (nesting available)!

        RETURN
        ------
        keys - all attr names (defined and not)
        values - Types!!! not instances!!!
        """
        result = {}
        for cls in self.iter_mro():
            _result_i = dict(cls.__annotations__)
            _result_i.update(result)
            result = _result_i
        return result

    def get__dict_values(self) -> TYPING.KWARGS_FINAL:
        """
        GOAL
        ----
        get dict with only existed values! no raise if value not exists!
        """
        result = {}
        for key in self.get__dict_types():
            if hasattr(self.SOURCE, key):
                result.update({key: getattr(self.SOURCE, key)})
        return result

    # -----------------------------------------------------------------------------------------------------------------
    def get__str_pretty(self) -> str:
        """just a pretty string for debugging or research.
        """
        result = f"{self.SOURCE.__class__.__name__}(Annotations):"
        annots = self.get__dict_values()
        if annots:
            for key, value in annots.items():
                result += f"\n\t{key}={value}"
        else:
            result += f"\nEmpty=Empty"

        print(result)
        return result

    # =================================================================================================================
    def iter_mro(self) -> Iterable[type]:
        """
        GOAL
        ----
        iter only important user classes from mro
        """
        yield from TypeAux(self.SOURCE).iter_mro_user(
            # NestGAI_AttrIC,
            # NestGSAI_AttrAnycase,
            # NestGA_AttrIC, NestGI_AttrIC,
            # NestSA_AttrAnycase, NestSI_AttrAnycase,
        )

    def iter_names(self) -> Iterable[str]:
        """
        iter all (with not existed)
        """
        yield from self.get__dict_types()

    def iter_values(self) -> Iterable[Any]:
        """
        only existed
        """
        yield from self.get__dict_values().values()

    # =================================================================================================================
    def values__set_none__existed(self) -> None:
        """
        GOAL
        ----
        set None for all annotated aux_attr! only existed!
        """
        for name in self.iter_names():
            if hasattr(self.SOURCE, name):
                setattr(self.SOURCE, name, None)

    def values__set_none__all(self) -> None:
        """
        GOAL
        ----
        set None for all annotated aux_attr! even not existed!
        """
        for name in self.iter_names():
            setattr(self.SOURCE, name, None)

    def values__delete(self) -> None:
        """
        GOAL
        ----
        delattr all annotated aux_attr!
        """
        for name in self.iter_names():
            if hasattr(self.SOURCE, name):
                delattr(self.SOURCE, name)

    def values__reinit_by_types(self, not_existed: bool = None) -> None:
        """
        GOAL
        ----
        delattr all annotated aux_attr!
        """
        for name, value in self.get__dict_types().items():
            if not_existed and hasattr(self.SOURCE, name):
                continue

            value = TypeAux(value).type__init_value__default()
            setattr(self.SOURCE, name, value)

    # -----------------------------------------------------------------------------------------------------------------
    def names__delete(self, *names: str) -> None:
        """
        ATTENTION
        ---------
        its not good idea!!! dont use it! yoг will change real CLASS parameters|

        GOAL
        ----
        del names from annots! if existed
        """
        if not names:
            names = self.iter_names()

        for cls in self.iter_mro():
            for name in names:
                if name in cls.__annotations__:
                    cls.__annotations__.pop(name)

    def names__add(self, *names: str) -> None:
        """
        ATTENTION
        ---------
        its not good idea!!! dont use it! yoг will change real CLASS parameters|

        GOAL
        ----
        add names into annots!
        """
        cls = TypeAux(self.SOURCE).get__class()

        for name in names:
            annots = cls.__annotations__
            if name not in annots:
                cls.__annotations__.update({name: Any})

    def name__get_original(self, name: str) -> str | None:
        """
        GOAL
        ----
        check name exists in annots ONLY!
        """
        annots = self.get__dict_types()
        name = IterAux(annots).item__get_original(name)
        if name == NoValue:
            name = None
        return name

    def name__check_exists(self, name: str) -> bool:
        """
        GOAL
        ----
        check name exists in annots ONLY!
        """
        return self.name__get_original(name) is not None


# =====================================================================================================================



import re

from base_aux.base_inits.m1_nest_init_source import *
from base_aux.base_statics.m4_enums import *
from base_aux.base_statics.m1_types import *
from base_aux.aux_callable.m1_callable_aux import CallableAux
from base_aux.aux_attr.m0_dump import AttrsDump


# =====================================================================================================================
@final
class AttrAux(NestInit_Source):
    """
    NOTE
    ----
    if there are several same aux_attr in different cases - you should resolve it by yourself!

    ANNOTS
    ------
    not used/intended here! dont mess they - it means you need to know what exact you will work with (attrs or annots)
    and prepare classes appropriate!
    """
    SOURCE: Any = AttrsDump

    # =================================================================================================================
    def get_name__private_external(self, dirname: str) -> str | None:
        """
        typically BUILTIN - NOT INCLUDED!

        NOTE
        ----
        BEST WAY TO USE EXACTLY iter__not_private

        GOAL
        ----
        using name (from dir(obj)) return user-friendly name! external name!

        REASON
        ------
        here in example - "__hello" will never appear directly!!!
        class Cls:
            ATTR1 = 1
            def __hello(self, *args) -> None:
                kwargs = dict.fromkeys(args)
                self.__init_kwargs(**kwargs)

        name='_Cls__hello' hasattr(self.SOURCE, name)=True
        name='__class__' hasattr(self.SOURCE, name)=True
        name='__delattr__' hasattr(self.SOURCE, name)=True
        name='__dict__' hasattr(self.SOURCE, name)=True
        name='__dir__' hasattr(self.SOURCE, name)=True
        name='__doc__' hasattr(self.SOURCE, name)=True
        ///
        name='ATTR1' hasattr(self.SOURCE, name)=True
        """
        # filter not hidden -------
        if not dirname.startswith("_"):
            return

        # filter private builtin -------
        if dirname.startswith("__"):
            return

        # parse private user -------
        if re.fullmatch(r"_.+__.+", dirname):
            # print(f"{dirname=}")
            # print(f"{self.SOURCE=}")
            try:
                # print(11)
                mro = self.SOURCE.__mro__
            except:
                # print(111)
                mro = self.SOURCE.__class__.__mro__
                # print(f"{mro=}")

            # fixme: cant solve problem for GetattrPrefixInst_RaiseIf! in case of _GETATTR__PREFIXES!!!
            for cls in mro:
                if dirname.startswith(f"_{cls.__name__}__"):
                    name_external = dirname.replace(f"_{cls.__name__}", "")
                    return name_external

    # =================================================================================================================
    # def __contains__(self, item: str):      # IN=DONT USE IT! USE DIRECT METHOD anycase__check_exists
    #     return self.anycase__check_exists(item)

    # ITER ------------------------------------------------------------------------------------------------------------
    def iter__external_not_builtin(self) -> Iterable[str]:
        """
        NOTE
        ----
        BEST WAY TO USE EXACTLY iter__not_private

        GOAL
        ----
        1/ iter only without builtins!!!
        2/ use EXT private names!

        SPECIALLY CREATED FOR
        ---------------------
        this class - all iterations!
        """
        for name in dir(self.SOURCE):
            # filter builtin ----------
            if name.startswith("__"):
                continue

            # filter private external ----------
            name_private_ext = self.get_name__private_external(name)
            if name_private_ext:
                yield name_private_ext
                continue

            # direct user attr ----------
            # print(f"{name=}")
            yield name

    # -----------------------------------------------------------------------------------------------------------------
    def iter__attrs(self, attr_level: AttrLevel = AttrLevel.NOT_PRIVATE) -> Iterable[str]:
        for name in self.iter__external_not_builtin():
            if attr_level == AttrLevel.NOT_PRIVATE:
                if not name.startswith("__"):
                    yield name

            elif attr_level == AttrLevel.NOT_HIDDEN:
                if not name.startswith("_"):
                    yield name

            elif attr_level == AttrLevel.PRIVATE:
                if name.startswith("__"):
                    yield name

            elif attr_level == AttrLevel.ALL:
                yield name

            else:
                raise Exx__Incompatible(f"{attr_level=}")

    def iter__not_hidden(self) -> Iterable[str]:
        """
        NOTE
        ----
        hidden names are more simple to detect then private!
        cause of private methods(!) changes to "_<ClsName><__MethName>"
        """
        return self.iter__attrs(AttrLevel.NOT_HIDDEN)

    def iter__not_private(self) -> Iterable[str]:
        """
        NOTE
        ----
        BEST WAY TO USE EXACTLY iter__not_private
        """
        return self.iter__attrs(AttrLevel.NOT_PRIVATE)

    def iter__private(self) -> Iterable[str]:
        """
        BUILTIN - NOT INCLUDED!

        NOTE
        ----
        BEST WAY TO USE EXACTLY iter__not_private

        GOAL
        ----
        collect all privates in original names! without ClassName-Prefix

        BEST IDEA
        ---------
        keep list of iters
        """
        return self.iter__attrs(AttrLevel.PRIVATE)


    # def __iter__(self):     # DONT USE IT! USE DIRECT METHODS
    #     yield from self.iter__not_hidden()

    # =================================================================================================================
    pass

    # NAME ------------------------------------------------------------------------------------------------------------
    def anycase__name_original(self, name: str) -> str | None:
        """
        get attr name in original register
        """
        if not isinstance(name, str):
            return

        name = str(name).strip()
        if not name:
            return

        for name_original in self.iter__external_not_builtin():
            if name_original.lower() == name.lower():
                return name_original

        return

    def anycase__check_exists(self, name: str) -> bool:
        return self.anycase__name_original(name) is not None

    # ATTR ------------------------------------------------------------------------------------------------------------
    def anycase__getattr(self, name: str) -> Any | Callable | NoReturn:
        """
        GOAL
        ----
        get attr value by name in any register
        no execution/resolving! return pure value as represented in object!
        """
        name_original = self.anycase__name_original(name)
        if name_original is None:
            raise AttributeError(name)

        value = getattr(self.SOURCE, name_original)
        return value

    def anycase__setattr(self, name: str, value: Any) -> None | NoReturn:
        """
        get attr value by name in any register
        no execution! return pure value as represented in object!

        NoReturn - in case of not accepted names when setattr
        """
        if not isinstance(name, str):
            raise TypeError(name)

        name = name.strip()
        if name in ["", ]:
            raise AttributeError(name)

        name_original = self.anycase__name_original(name)
        if name_original is None:
            name_original = name

        # NOTE: you still have no exx with setattr(self.SOURCE, "    HELLO", value) and ""
        setattr(self.SOURCE, name_original, value)

    def anycase__delattr(self, name: str) -> None:
        name_original = self.anycase__name_original(name)
        if name_original is None:
            return      # already not exists

        delattr(self.SOURCE, name_original)

    # ITEM ------------------------------------------------------------------------------------------------------------
    def anycase__getitem(self, name: str) -> Any | Callable | NoReturn:
        return self.anycase__getattr(name)

    def anycase__setitem(self, name: str, value: Any) -> None | NoReturn:
        self.anycase__setattr(name, value)

    def anycase__delitem(self, name: str) -> None:
        self.anycase__delattr(name)

    # =================================================================================================================
    def getattr__callable_resolve(self, name: str, callables_resolve: CallableResolve = CallableResolve.DIRECT) -> Any | Callable | CallableResolve | NoReturn:
        """
        SAME AS
        -------
        CallableAux(*).resolve_*

        WHY NOT-1=CallableAux(*).resolve_*
        ----------------------------------
        it is really the same, BUT
        1. additional try for properties (could be raised without calling)
        2. cant use here cause of Circular import accused
        """
        # resolve property --------------
        # result_property = CallableAux(getattr).resolve(callables_resolve, self.SOURCE, realname)
        # TypeAux

        try:
            value = self.anycase__getattr(name)
        except Exception as exx:
            if callables_resolve == CallableResolve.SKIP_RAISED:
                return ProcessState.SKIPPED
            elif callables_resolve == CallableResolve.EXX:
                return exx
            elif callables_resolve == CallableResolve.RAISE_AS_NONE:
                return None
            elif callables_resolve == CallableResolve.RAISE:
                raise exx
            elif callables_resolve == CallableResolve.BOOL:
                return False
            else:
                raise exx

        # resolve callables ------------------
        result = CallableAux(value).resolve(callables_resolve)
        return result

    # =================================================================================================================
    def load__by_dict(self, other: dict[str, Any], only_existed: bool = None) -> Any | AttrsDump:
        """
        MAIN ITEA
        ----------
        LOAD MEANS basically setup final values for final not callables values!
        but you can use any types for your own!
        expected you know what you do and do exactly ready to use final values/not callables in otherObj!

        NOTE
        ----
        1/ dont use callables_resolve here
        2/ dont add any TEMPLATES!!! its responsibility for DictAux! use it by yourself!!!
        """
        # work ----------
        for key, value in other.items():
            if only_existed and not self.anycase__check_exists(key):
                continue
            self.anycase__setattr(key, value)

        return self.SOURCE

    # DUMP ------------------------------------------------------------------------------------------------------------
    def dump_dict(self, callables_resolve: CallableResolve = CallableResolve.EXX) -> dict[str, Any | Callable | Exception] | NoReturn:
        """
        MAIN IDEA
        ----------
        BUMPS MEANS basically save final values for all (even any dynamic/callables) values! or only not callables!

        NOTE
        ----
        DUMP WITHOUT PRIVATE NAMES

        GOAL
        ----
        make a dict from any object from aux_attr (not hidden)

        SPECIALLY CREATED FOR
        ---------------------
        using any object as rules for Translator
        """
        result = {}
        for name in self.iter__not_private():
            value = self.getattr__callable_resolve(name=name, callables_resolve=callables_resolve)
            if value == ProcessState.SKIPPED:
                continue
            result.update({name: value})

        return result

    def dump_dict__resolve_exx(self) -> dict[str, Any | Exception]:
        """
        MAIN DERIVATIVE!
        """
        return self.dump_dict(CallableResolve.EXX)

    def dump_dict__direct(self) -> TYPING.KWARGS_FINAL:
        return self.dump_dict(CallableResolve.DIRECT)

    def dump_dict__skip_callables(self) -> TYPING.KWARGS_FINAL:
        return self.dump_dict(CallableResolve.SKIP_CALLABLE)

    def dump_dict__skip_raised(self) -> dict[str, Any] | NoReturn:
        return self.dump_dict(CallableResolve.RAISE)

    # -----------------------------------------------------------------------------------------------------------------
    def dump_obj(self, callables_resolve: CallableResolve = CallableResolve.EXX) -> AttrsDump | NoReturn:
        data = self.dump_dict(callables_resolve)
        obj = AttrAux().load__by_dict(data)
        return obj

    # -----------------------------------------------------------------------------------------------------------------
    def dump__pretty_str(self) -> str:
        result = f"{self.SOURCE.__class__.__name__}(Attributes):"
        for key, value in self.dump_dict(CallableResolve.EXX).items():
            result += f"\n\t{key}={value}"
        else:
            result += f"\nEmpty=Empty"

        return result


# =====================================================================================================================



from base_aux.aux_callable.m1_callable_aux import *
from base_aux.base_statics.m1_types import *


# =====================================================================================================================
class _EqValidator:
    """
    MAIN IDEA
    ---------
    ALL WHAT PASSED INTO INIT WOULD PASS INTO VALIDATOR() AFTER FIRST ARG (OTHER)

    NOTE
    ----
    1/ preferably not use directly this object!
    USE DERIVATIVES!!! without validator passing

    2/ MAIN IDEA - NEVER RAISED!!! if any - return FALSE!!! if need - check manually!
    why so? - because i need smth to make a tests with final result of any source!
    dont mind reason!

    GOAL
    ----
    base object to make a validation by direct comparing with other object
    no raise

    USAGE
    -----
    PLACE ONLY IN FIRST PLACE!
    """
    VALIDATOR: TYPE__VALID_VALIDATOR

    V_ARGS: TYPING.ARGS_FINAL
    V_KWARGS: TYPING.KWARGS_FINAL

    REVERSE: bool = None

    OTHER_RAISED: bool = None
    OTHER_FINAL: Any | Exception = None

    def __init__(self, validator: TYPE__VALID_VALIDATOR, *v_args, reverse: bool = None, **v_kwargs) -> None:
        if validator is not None:
            self.VALIDATOR = validator

        if reverse is not None:
            self.REVERSE = reverse

        # super(ArgsKwargs, self).__init__(*v_args, **v_kwargs)
        self.V_ARGS = v_args
        self.V_KWARGS = v_kwargs

    def __str__(self):
        args = self.V_ARGS
        kwargs = self.V_KWARGS
        reverse = self.REVERSE
        return f"{self.__class__.__name__}({args=},{kwargs=},{reverse=})"

    def __repr__(self):
        """
        GOAL
        ----
        used in
        """
        return str(self)

    def __eq__(self, other_draft) -> bool:
        return self.validate(other_draft)

    def __call__(self, other_draft: Any, *other_args, **other_kwargs) -> bool:
        """
        NOTE
        ----
        other_args/* - only for manual usage!
        typically used only other and only by direct eq(o1 == o2)
        """
        return self.validate(other_draft, *other_args, **other_kwargs)

    def __contains__(self, item) -> bool:
        return self.validate(item)

    def __iter__(self) -> Iterable[Any]:
        """
        NOTE
        ----
        not always correct!
        best usage for EqVariants or for any object with several args (Reqexp/...)
        """
        yield from self.V_ARGS

    def validate(self, other_draft: Any, *other_args, **other_kwargs) -> bool:
        """
        GOAL
        ----
        validate smth with special logic
        """
        # ------
        # TODO: decide use or not callable other??? = USE! it is really need to validate callable!!!
        try:
            self.OTHER_FINAL = CallableAux(other_draft).resolve_raise(*other_args, **other_kwargs)
            self.OTHER_RAISED = False
        except Exception as exx:
            self.OTHER_RAISED = True
            self.OTHER_FINAL = exx

        result = CallableAux(self.VALIDATOR).resolve_bool(self.OTHER_FINAL, *self.V_ARGS, **self.V_KWARGS)
        if self.REVERSE:
            result = not result
        return result

    def VALIDATOR(self, other_final, *v_args, **v_kwargs) -> bool | NoReturn:
        return NotImplemented


# =====================================================================================================================
class EqValid_Base(_EqValidator):
    def __init__(self, *v_args, **v_kwargs):
        # print(v_args, v_kwargs)
        # super(ArgsKwargs, self).__init__(*v_args, **v_kwargs)     # not working!

        super().__init__(None, *v_args, **v_kwargs)
        # self.V_ARGS = v_args
        # self.V_KWARGS = v_kwargs
        #
        # if reverse is not None:
        #     self.REVERSE = reverse


# =====================================================================================================================



from enum import Enum


# =====================================================================================================================
# TODO: make own class! EnumIc with metaclass!
class NestEq_Enum(Enum):
    """
    # NOTE
    # ----
    # DEL=work IC only on Eq! not working with Contains and Init!!! need edit Metaclass!

    GOAL
    ----
    add user friendly cmp objects with final values

    VictimEq(1) == 1    # for std object it is False
    """
    # TODO: add Contain classmeth???  cant understand! need metaclass!

    # def __eq__(self, other) -> bool:
    #     if isinstance(other, self.__class__):
    #         return self.value == other.value        # or str(self.value).lower() == str(other.value).lower()    # NO!!!
    #
    #     else:
    #         for enum_i in self.__class__:
    #             if isinstance(other, str):
    #                 if str(enum_i.value).lower() == str(other).lower():
    #                     return True
    #             else:
    #                 try:
    #                     if enum_i.value == other or other == enum_i.value:
    #                         return True
    #                 except:
    #                     pass
    #     return False

    def __eq__(self, other) -> bool:
        result = False
        if isinstance(other, self.__class__):
            return self.value == other.value

        if other in self.__class__:
            result = True
        try:
            if other in self:
                result = True
        except:
            pass

        if result:
            return other == self.value or self == self.__class__(other)
        else:
            return False

    # @classmethod
    # def __contains__(cls, other) -> bool:
    #     if isinstance(other, cls):
    #         other = other.value
    #
    #     for enum_i in cls:
    #         if enum_i.value == other or str(enum_i.value).lower() == str(other).lower():
    #             return True
    #     else:
    #         return False


# =====================================================================================================================



from typing import *
import pathlib
from base_aux.aux_values.m0_novalue import *


# =====================================================================================================================
class _Cls:
    def meth(self):
        pass
    @property
    def prop(self):
        return


def _explore():
    print(isinstance(_Cls.prop, property))  # True
    print(type(_Cls.prop))  # <class 'property'>
    print(type(_Cls().prop))  # <class 'NoneType'>


# =====================================================================================================================
OBJ_NAMES__BUILTINS = tuple(dir(globals().get("__builtins__")))
"""
('ArithmeticError', 'AssertionError', 'AttributeError', 'BaseException', 'BaseExceptionGroup', 'BlockingIOError', 'BrokenPipeError', 'BufferError', 'BytesWarning', 'ChildProcessError', 'ConnectionAbortedError', 'ConnectionError', 'ConnectionRefusedError', 'ConnectionResetError', 'DeprecationWarning', 'EOFError', 'Ellipsis', 'EncodingWarning', 'EnvironmentError', 'Exception', 'ExceptionGroup', 'False', 'FileExistsError', 'FileNotFoundError', 'FloatingPointError', 'FutureWarning', 'GeneratorExit', 'IOError', 'ImportError', 'ImportWarning', 'IndentationError', 'IndexError', 'InterruptedError', 'IsADirectoryError', 'KeyError', 'KeyboardInterrupt', 'LookupError', 'MemoryError', 'ModuleNotFoundError', 'NameError', 'None', 'NotADirectoryError', 'NotImplemented', 'NotImplementedError', 'OSError', 'OverflowError', 'PendingDeprecationWarning', 'PermissionError', 'ProcessLookupError', 'RecursionError', 'ReferenceError', 'ResourceWarning', 'RuntimeError', 'RuntimeWarning', 'StopAsyncIteration', 'StopIteration', 'SyntaxError', 'SyntaxWarning', 'SystemError', 'SystemExit', 'TabError', 'TimeoutError', 'True', 'TypeError', 'UnboundLocalError', 'UnicodeDecodeError', 'UnicodeEncodeError', 'UnicodeError', 'UnicodeTranslateError', 'UnicodeWarning', 'UserWarning', 'ValueError', 'Warning', 'WindowsError', 'ZeroDivisionError', '__build_class__', '__debug__', '__doc__', '__import__', '__loader__', '__name__', '__package__', '__spec__', 'abs', 'aiter', 'all', 'anext', 'any', 'ascii', 'bin', 'bool', 'breakpoint', 'bytearray', 'bytes', 'callable', 'chr', 'classmethod', 'compile', 'complex', 'copyright', 'credits', 'delattr', 'dict', 'dir', 'divmod', 'enumerate', 'eval', 'exec', 'exit', 'filter', 'float', 'format', 'frozenset', 'getattr', 'globals', 'hasattr', 'hash', 'help', 'hex', 'id', 'input', 'int', 'isinstance', 'issubclass', 'iter', 'len', 'license', 'list', 'locals', 'map', 'max', 'memoryview', 'min', 'next', 'object', 'oct', 'open', 'ord', 'pow', 'print', 'property', 'quit', 'range', 'repr', 'reversed', 'round', 'set', 'setattr', 'slice', 'sorted', 'staticmethod', 'str', 'sum', 'super', 'tuple', 'type', 'vars', 'zip')
"""


# COLLECTIONS =========================================================================================================
@final
class TYPES:
    """
    GOAL
    ----
    collect all types USEFUL variants
    """
    # SINGLE ---------------------------
    NONE: type = type(None)
    NUMBER = int | float

    FUNCTION: type = type(lambda: True)
    METHOD: type = type(_Cls().meth)
    # PROPERTY: type    # cant check by type
    # MODULE: type    # cant check by type - only by method!

    # COLLECTIONS ---------------------------
    ELEMENTARY_SINGLE: tuple[type, ...] = (
        type(None),
        bool,
        int, float,
        str, bytes,
    )
    ELEMENTARY_COLLECTION: tuple[type, ...] = (
        tuple, list,
        set, dict,
    )
    ELEMENTARY: tuple[type, ...] = (
        *ELEMENTARY_SINGLE,
        *ELEMENTARY_COLLECTION,
    )


# =====================================================================================================================
@final
class TYPING:
    """
    GOAL
    ----
    collect all typing USER variants
    """
    ELEMENTARY = Union[*TYPES.ELEMENTARY]

    # -----------------------------------------------------------------------------------------------------------------
    ARGS_FINAL = tuple[Any, ...]
    ARGS_DRAFT = Union[Any, ARGS_FINAL, 'ArgsKwargs']           # you can use direct single value

    KWARGS_FINAL = dict[str, Any]
    KWARGS_DRAFT = Union[None, KWARGS_FINAL, 'ArgsKwargs']  # if passed NONE - no data!

    # -----------------------------------------------------------------------------------------------------------------
    PATH_FINAL = pathlib.Path
    PATH_DRAFT = Union[str, PATH_FINAL]

    STR_FINAL = str
    STR_DRAFT = Union[STR_FINAL, Any]

    # -----------------------------------------------------------------------------------------------------------------
    DICT_ANY_ANY = dict[Any, Any]               # just to show - dict could be any! on keys/values
    DICT_STR_ANY = dict[str, Any]               # just to show - dict could be any! on values! not just an elementary1
    DICT_STR_ELEM = DICT_JSON_ANY = dict[str, ELEMENTARY]       # just to show - parsed by json - dict
    DICT_STR_STR = DICT_INI = dict[str, str]               # just to show - parsed by ini!
    JSON_ANY = ELEMENTARY | DICT_STR_ELEM  # just to show - parsed by json - any object

    # -----------------------------------------------------------------------------------------------------------------
    ORDERED_ITERABLE = Union[dict, list, tuple, Iterable]     # "SET" - DONT USE!
    ITERPATH_KEY = Union[Any, int]   # Any is for dict
    ITERPATH = tuple[ITERPATH_KEY, ...]


# VALUES --------------------------------------------------------------------------------------------------------------
ARGS_FINAL__BLANK = ()
KWARGS_FINAL__BLANK = {}


# VALIDS ==============================================================================================================
TYPE__VALID_EXX = Union[Exception, type[Exception]]
TYPE__VALID_RESULT = Union[
    Any,
    TYPE__VALID_EXX,  # as main idea! instead of raise
]
TYPE__VALID_BOOL__DRAFT = Union[
    Any,                                # fixme: hide? does it need? for results like []/{}/()/0/"" think KEEP! it mean you must know that its expecting boolComparing in further logic!
    bool,                               # as main idea! as already final generic
    Callable[[...], bool | Any | NoReturn],   # as main idea! to get final generic
    TYPE__VALID_EXX,
    NoValue
]
TYPE__VALID_BOOL__FINAL = Union[
    # this is when you need get only bool! raise - as False!
    bool,  # as main idea! instead of raise/exx
]
TYPE__VALID_BOOL_EXX__FINAL = Union[
    bool,
    TYPE__VALID_EXX,
]
# TYPE__VALID_TRUE__FINAL = Union[

TYPE__VALID_VALIDATOR = Union[
    Any,    # generic final instance as expecting value - direct comparison OR comparison instance like Valid!
    # Type,   # Class as validator like Exception????? fixme
    type[Exception],  # direct comparison
    Callable[[Any, ...], bool | NoReturn]     # func with first param for validating source
]


# =====================================================================================================================
TYPE__LAMBDA_CONSTRUCTOR = Union[Any, type[Any], Callable[..., Any | NoReturn]]
_TYPE__LAMBDA_BOOL = Union[Any, Callable[..., bool | NoReturn]]


# =====================================================================================================================



from typing import *


# =====================================================================================================================
class NestCmp:
    """
    GOAL
    ----
    APPLYING COMPARISON WITH SELF INSTANCE

    BEST USAGE
    ----------
    just redefine one method __cmp__!

    WHY NOT: JUST USING ONE BY ONE EXACT METHODS?
    ---------------------------------------------
    it is more complicated then just one explicit __cmp__()!
    __cmp__ is not directly acceptable in Python! this is not a buildIn method!

    """
    __eq__ = lambda self, other: self.__cmp__(other) == 0
    # __ne__ = lambda self, other: self.__cmp__(other) != 0

    __lt__ = lambda self, other: self.__cmp__(other) < 0
    __gt__ = lambda self, other: self.__cmp__(other) > 0
    __le__ = lambda self, other: self.__cmp__(other) <= 0
    __ge__ = lambda self, other: self.__cmp__(other) >= 0

    ltgt = lambda self, other1, other2: self > other1 and self < other2
    ltge = lambda self, other1, other2: self > other1 and self <= other2

    legt = lambda self, other1, other2: self >= other1 and self < other2
    lege = lambda self, other1, other2: self >= other1 and self <= other2

    # CMP -------------------------------------------------------------------------------------------------------------
    def __cmp__(self, other: Any) -> int | NoReturn:
        """
        do try to resolve Exceptions!!! sometimes it is ok to get it!!!

        RETURN
        ------
            1=self>other
            0=self==other
            -1=self<other
        """
        raise NotImplemented()

    # -----------------------------------------------------------------------------------------------------------------
    # def __eq__(self, other):
    #     return self.__cmp__(other) == 0
    #
    # def __ne__(self, other):
    #     return self.__cmp__(other) != 0
    #
    # def __lt__(self, other):
    #     return self.__cmp__(other) < 0
    #
    # def __gt__(self, other):
    #     return self.__cmp__(other) > 0
    #
    # def __le__(self, other):
    #     return self.__cmp__(other) <= 0
    #
    # def __ge__(self, other):
    #     return self.__cmp__(other) >= 0


# =====================================================================================================================



from typing import *

from base_aux.base_statics.m4_enums import *
from base_aux.valid.m1_aux_valid_lg import *
from base_aux.aux_types.m1_type_aux import *
from base_aux.aux_attr.m1_attr1_aux import *
from base_aux.aux_cmp_eq.m2_eq_aux import *
from base_aux.aux_attr.m2_annot1_aux import *


# =====================================================================================================================
class Validators:
    """
    GOAL
    ----
    collect all validators (funcs) in one place
    applicable in EqValid_Base only (by common way), but you can try using it separated!

    SPECIALLY CREATED FOR
    ---------------------
    EqValid_Base

    RULES
    -----
    1/ NoReturn - available for all returns as common!!! but sometimes it cant be reached (like TRUE/RAISE)
    """
    def VariantsDirect(self, other_final: Any, *variants: Any) -> bool | NoReturn:
        return other_final in variants

    def VariantsStrLow(self, other_final: Any, *variants: Any) -> bool | NoReturn:
        other_final = str(other_final).lower()
        variants = (str(var).lower() for var in variants)

        return other_final in variants

    # -----------------------------------------------------------------------------------------------------------------
    def Isinstance(self, other_final: Any, *variants: type[Any]) -> bool | NoReturn:
        for variant in variants:
            if isinstance(other_final, variant):
                return True
        return False

    # -----------------------------------------------------------------------------------------------------------------
    def Startswith(self, other_final: Any, *variants: Any, ignorecase: bool = None) -> bool | NoReturn:
        if ignorecase:
            other_final = str(other_final).lower()
            variants = (str(var).lower() for var in variants)
        else:
            other_final = str(other_final)
            variants = (str(_) for _ in variants)

        for var in variants:
            if other_final.startswith(var):
                return True

        return False

    def Endswith(self, other_final: Any, *variants: Any, ignorecase: bool = None) -> bool | NoReturn:
        if ignorecase:
            other_final = str(other_final).lower()
            variants = (str(var).lower() for var in variants)
        else:
            other_final = str(other_final)
            variants = (str(_) for _ in variants)

        for var in variants:
            if other_final.endswith(var):
                return True

        return False

    # -----------------------------------------------------------------------------------------------------------------
    def TRUE(self, other_final: TYPE__VALID_BOOL__DRAFT, *v_args, **v_kwargs) -> bool:
        """
        GOAL
        ----
        True - if Other object called with no raise and no Exception in result
        """
        result = False
        if self.OTHER_RAISED or TypeAux(other_final).check__exception():
            return False

        return bool(other_final)

    # TODO: add FALSE????? what to do with exx and real false?

    def Raise(self, other_final: Any, *variants: Any) -> bool:
        """
        GOAL
        ----
        True - if Other object called with raised
        if other is exact final Exception without raising - it would return False!
        """
        return self.OTHER_RAISED

    def NotRaise(self, other_final, *v_args, **v_kwargs) -> bool:
        """
        GOAL
        ----
        True - if Other object called with raised
        if other is exact final Exception without raising - it would return False!
        """
        return not self.OTHER_RAISED

    def Exx(self, other_final, *v_args, **v_kwargs) -> bool:
        """
        GOAL
        ----
        True - if Other object is exact Exception or Exception()
        if raised - return False!!
        """
        return not self.OTHER_RAISED and TypeAux(other_final).check__exception()

    def ExxRaise(self, other_final, *v_args, **v_kwargs) -> bool:
        """
        GOAL
        ----
        True - if Other object is exact Exception or Exception() or Raised
        """
        return self.OTHER_RAISED or TypeAux(other_final).check__exception()

    # -----------------------------------------------------------------------------------------------------------------
    def LtGt_Obj(self, other_final, low: Any | None = None, high: Any | None = None) -> bool | NoReturn:
        return ValidAux_Obj(other_final).ltgt(low, high)

    def LtGe_Obj(self, other_final, low: Any | None = None, high: Any | None = None) -> bool | NoReturn:
        return ValidAux_Obj(other_final).ltge(low, high)

    def LeGt_Obj(self, other_final, low: Any | None = None, high: Any | None = None) -> bool | NoReturn:
        return ValidAux_Obj(other_final).legt(low, high)

    def LeGe_Obj(self, other_final, low: Any | None = None, high: Any | None = None) -> bool | NoReturn:
        return ValidAux_Obj(other_final).lege(low, high)

    # -----------------------------------------------------------------------------------------------------------------
    def LtGt_NumParsedSingle(self, other_final, low: Any | None = None, high: Any | None = None) -> bool | NoReturn:
        return ValidAux_NumParsedSingle(other_final).ltgt(low, high)

    def LtGe_NumParsedSingle(self, other_final, low: Any | None = None, high: Any | None = None) -> bool | NoReturn:
        return ValidAux_NumParsedSingle(other_final).ltge(low, high)

    def LeGt_NumParsedSingle(self, other_final, low: Any | None = None, high: Any | None = None) -> bool | NoReturn:
        return ValidAux_NumParsedSingle(other_final).legt(low, high)

    def LeGe_NumParsedSingle(self, other_final, low: Any | None = None, high: Any | None = None) -> bool | NoReturn:
        return ValidAux_NumParsedSingle(other_final).lege(low, high)

    # -----------------------------------------------------------------------------------------------------------------
    def NumParsedSingle(self, other_final, expect: Any | None | bool | NumType = True) -> bool:
        return ValidAux_NumParsedSingle(other_final).eq(expect)

    def NumParsedSingle_Int(self, other_final) -> bool:
        return ValidAux_NumParsedSingle(other_final).eq(int)

    def NumParsedSingle_Float(self, other_final) -> bool:
        return ValidAux_NumParsedSingle(other_final).eq(float)

    # -----------------------------------------------------------------------------------------------------------------
    def Regexp(
            self,
            other_final,
            *regexps: str,
            ignorecase: bool = True,
            bool_collect: BoolCumulate = None,
            match_link: Callable = re.fullmatch,
    ) -> bool | NoReturn:
        bool_collect = bool_collect or self.BOOL_COLLECT

        for pattern in regexps:
            result_i = match_link(pattern=str(pattern), string=str(other_final), flags=re.RegexFlag.IGNORECASE if ignorecase else 0)

            # CUMULATE --------
            if bool_collect == BoolCumulate.ALL_TRUE:
                if not result_i:
                    return False
            elif bool_collect == BoolCumulate.ANY_TRUE:
                if result_i:
                    return True
            elif bool_collect == BoolCumulate.ALL_FALSE:
                if result_i:
                    return False
            elif bool_collect == BoolCumulate.ANY_FALSE:
                if not result_i:
                    return True

        # FINAL ------------
        if bool_collect in [BoolCumulate.ALL_TRUE, BoolCumulate.ALL_FALSE]:
            return True
        else:
            return False

    # -----------------------------------------------------------------------------------------------------------------
    def AttrsByKwargs(
            self,
            other_final,
            # callable_resolve: CallableResolve = CallableResolve.EXX,
            **kwargs: TYPING.KWARGS_FINAL
    ) -> bool | NoReturn:
        for key, value in kwargs.items():
            value_expected = CallableAux(value).resolve(CallableResolve.EXX)
            value_other = AttrAux(other_final).getattr__callable_resolve(key, CallableResolve.EXX)
            if not EqAux(value_expected).check_doubleside__bool(value_other):
                return False

        # FINISH -----
        return True

    def AttrsByObj(
            self,
            other_final,
            # callable_resolve: CallableResolve = CallableResolve.EXX,
            source: Any,
            # attr_level: AttrLevel = AttrLevel.NOT_PRIVATE,
    ) -> bool | NoReturn:
        for key in AttrAux(source).iter__attrs(self.ATTR_LEVEL):
            value_expected = AttrAux(source).getattr__callable_resolve(key, CallableResolve.EXX)
            value_other = AttrAux(other_final).getattr__callable_resolve(key, CallableResolve.EXX)
            if not EqAux(value_expected).check_doubleside__bool(value_other):
                return False

        # FINISH -----
        return True

    # NOTE: INAPPROPRIATE!!!!
    # def AttrsByObjNotPrivate(
    #         self,
    #         other_final,
    #         # callable_resolve: CallableResolve = CallableResolve.EXX,
    #         source: Any,
    # ) -> bool | NoReturn:
    #     return self._AttrsByObj(other_final=other_final, source=source, attr_level=AttrLevel.NOT_PRIVATE)
    # def AttrsByObjNotHidden(
    #         self,
    #         other_final,
    #         # callable_resolve: CallableResolve = CallableResolve.EXX,
    #         source: Any,
    # ) -> bool | NoReturn:
    #     return self._AttrsByObj(other_final=other_final, source=source, attr_level=AttrLevel.NOT_HIDDEN)

    # -----------------------------------------------------------------------------------------------------------------
    def AnnotsAllExists(
            self,
            other_final,
            **kwargs: TYPING.KWARGS_FINAL
    ) -> bool | NoReturn:
        return AnnotsAux(other_final).check_all_defined()


# =====================================================================================================================



from typing import *
import smtplib
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart

from base_aux.alerts.m0_base import *
from base_aux.privates.m1_privates import *
from base_aux.aux_attr.m4_kits import *
from base_aux.aux_attr.m1_attr2_nest2_lambdas_resolve import NestInit_AttrsLambdaResolve


# =====================================================================================================================
class SmtpAddress(NamedTuple):
    """class for keeping connection parameters/settings for exact smtp server

    :ivar ADDR: smtp server address like "smtp.mail.ru"
    :ivar PORT: smtp server port like 465
    """
    ADDR: str
    PORT: int


class SmtpServers:
    """well known servers addresses.

    Here we must collect servers like MilRu/GmailCom, and not to create it in any new project.
    """
    MAIL_RU: SmtpAddress = SmtpAddress("smtp.mail.ru", 465)


# =====================================================================================================================
class AlertSmtp(NestInit_AttrsLambdaResolve, AlertBase):
    """SMTP realisation for sending msg (email).
    """
    # SETTINGS ------------------------------------
    SERVER_SMTP: SmtpAddress = SmtpServers.MAIL_RU
    AUTH: AttrKit_AuthNamePwd = PvLoaderIni_AuthNamePwd(keypath=("AUTH_EMAIL_DEF",))
    TIMEOUT_SEND = 5

    # AUX -----------------------------------------
    _conn:  smtplib.SMTP_SSL

    def _connect_unsafe(self) -> Union[bool, NoReturn]:
        self._conn = smtplib.SMTP_SSL(self.SERVER_SMTP.ADDR, self.SERVER_SMTP.PORT, timeout=self.TIMEOUT_SEND)
        return True

    def _login_unsafe(self) -> Union[bool, NoReturn]:
        response = self._conn.login(self.AUTH.NAME, self.AUTH.PWD)
        print(response)
        print("=" * 100)
        return response and response[0] in [235, 503]

    def _send_unsafe(self) -> Union[bool, NoReturn]:
        self._conn.send_message(self._msg_compose())
        return True

    def _msg_compose(self) -> MIMEMultipart:
        msg = MIMEMultipart()
        msg["From"] = self.AUTH.NAME
        msg["To"] = self.RECIPIENT_SPECIAL or self.AUTH.NAME
        msg['Subject'] = self.SUBJECT
        msg.attach(MIMEText(self.body, _subtype=self._subtype))
        return msg

    def _recipient_self_get(self) -> str:
        return self.AUTH.NAME


# =====================================================================================================================



from typing import *
from dataclasses import dataclass, field

from .m1_type_aux import TypeAux
from base_aux.aux_attr.m0_static import check_name__buildin


# =====================================================================================================================
def _value_search_by_list(source: Any, search_list: list[Any]) -> Any | None:
    search_list = search_list or []
    for search_item in search_list:
        try:
            if search_item in source:
                return search_item
        except:
            pass

        try:
            if search_item == source:
                return search_item
        except:
            pass


# =====================================================================================================================
class ItemInternal(NamedTuple):
    KEY: str
    VALUE: str


# =====================================================================================================================
@dataclass
class ObjectState:
    """
    class for keeping results
    """
    # TODO: add sort method!!!
    SKIPPED_FULLNAMES: list[str] = field(default_factory=list)
    SKIPPED_PARTNAMES: list[str] = field(default_factory=list)

    PROPERTIES__ELEMENTARY_SINGLE: dict[str, Any] = field(default_factory=dict)
    PROPERTIES__ELEMENTARY_COLLECTION: dict[str, Any] = field(default_factory=dict)
    PROPERTIES__OBJECTS: dict[str, Any] = field(default_factory=dict)
    PROPERTIES__EXX: dict[str, Exception] = field(default_factory=dict)

    METHODS__ELEMENTARY_SINGLE: dict[str, Any] = field(default_factory=dict)
    METHODS__ELEMENTARY_COLLECTION: dict[str, Any] = field(default_factory=dict)
    METHODS__OBJECTS: dict[str, Any] = field(default_factory=dict)
    METHODS__EXX: dict[str, Exception] = field(default_factory=dict)


# =====================================================================================================================
class ObjectInfo:
    """
    :ivar MAX_ITER_ITEMS: 0 or None if not limited!
    """
    # SETTINGS --------------------------------------------
    MAX_LINE_LEN: int = 100
    MAX_ITER_ITEMS: int = 5
    HIDE_BUILD_IN: bool = None
    LOG_ITER: bool = None

    NAMES__USE_ONLY_PARTS: list[str] = []
    NAMES__SKIP_FULL: list[str] = [
    ]
    NAMES__SKIP_PARTS: list[str] = [
        # DANGER
        "init", "new", "create", "enter", "install",
        "set",
        "clone", "copy", "move",
        "next",
        "clear", "reduce",
        "close", "del", "exit", "kill", "abort",

        # PyQt5 Qthread
        "exec", "exec_", "pyqtConfigure",
        "dump",     # 'dumpObjectInfo' from PyQt5.QMenu

        # GIT
        "checkout", "detach",

        # threads
        "run", "start", "wait", "join", "terminate", "quit", "disconnect",

        # change collection content/count/order
        "pop", "popleft",
        "append", "appendleft",
        "extend", "extendleft",
        "add", "insert",
        "reverse", "rotate", "sort",
    ]

    # AUX --------------------------------------------------
    ITEM_CLS: type[ObjectState] = ObjectState
    ITEM: ObjectState = ITEM_CLS()
    SOURCE: Any = None

    def __init__(
            self,
            source: Optional[Any] = None,

            max_line_len: Optional[int] = None,
            max_iter_items: Optional[int] = None,
            hide_build_in: Optional[bool] = None,
            log_iter: Optional[bool] = None,

            names__use_only_parts: Union[None, str, list[str]] = None,
            names__skip_full: Union[None, str, list[str]] = None,
            names__skip_parts: Union[None, str, list[str]] = None,
    ):

        # SETTINGS -----------------------------------------------------------
        # RAPAMS -----------------------
        if max_line_len is not None:
            self.MAX_LINE_LEN = max_line_len
        if max_iter_items is not None:
            self.MAX_ITER_ITEMS = max_iter_items
        if hide_build_in is not None:
            self.HIDE_BUILD_IN = hide_build_in
        if log_iter is not None:
            self.LOG_ITER = log_iter

        # LISTS -----------------------
        if names__use_only_parts:
            if isinstance(names__use_only_parts, str):
                names__use_only_parts = [names__use_only_parts, ]
            self.NAMES__USE_ONLY_PARTS = names__use_only_parts
        if names__skip_full:
            if isinstance(names__skip_full, str):
                names__skip_full = [names__skip_full, ]
            self.NAMES__SKIP_FULL.extend(names__skip_full)
        if names__skip_parts:
            if isinstance(names__skip_parts, str):
                names__skip_full = [names__skip_parts, ]
            self.NAMES__SKIP_PARTS.extend(names__skip_parts)

        # WORK -----------------------------------------------------------
        self._item_clear()

        if source is not None:
            self.SOURCE = source

    # =================================================================================================================
    def _item_clear(self) -> None:
        self.ITEM = self.ITEM_CLS()

    def _item_reload(self) -> None:
        self._item_clear()

        # WORK --------------------------------------
        name = f"log_iter={self.LOG_ITER}(wait last touched)"
        self._print_line__group_separator(name)

        for pos, name in enumerate(dir(self.SOURCE), start=1):
            if self.LOG_ITER:
                print(f"{pos}:\t{name}")

            # FILTER -----------------------------------------------------------------------
            if self.HIDE_BUILD_IN and check_name__buildin(name):
                continue

            if self.NAMES__USE_ONLY_PARTS:
                use_name = False
                for name_include_item in self.NAMES__USE_ONLY_PARTS:
                    if name_include_item.lower() in name.lower():
                        use_name = True
                        break
                if not use_name:
                    continue

            # SKIP ------------------------------------------------------------------------
            SKIP_FULLNAMES = []
            if self.NAMES__SKIP_FULL:
                SKIP_FULLNAMES.extend(self.NAMES__SKIP_FULL)
            if SKIP_FULLNAMES:
                SKIP_FULLNAMES.extend(SKIP_FULLNAMES)
            if name in SKIP_FULLNAMES:
                self.ITEM.SKIPPED_FULLNAMES.append(name)
                continue

            SKIP_PARTNAMES = []
            if self.NAMES__SKIP_PARTS:
                SKIP_PARTNAMES.extend(self.NAMES__SKIP_PARTS)
            if SKIP_PARTNAMES:
                SKIP_PARTNAMES.extend(SKIP_PARTNAMES)
            if _value_search_by_list(source=name, search_list=SKIP_PARTNAMES):
                self.ITEM.SKIPPED_PARTNAMES.append(name)
                continue

            # PROPERTIES/METHODS + Exception--------------------------------------------------
            attr_is_method: bool = False
            try:
                value = getattr(self.SOURCE, name)
            except Exception as exx:
                self.ITEM.PROPERTIES__EXX.update({name: exx})
                continue

            if callable(value):
                attr_is_method = True
                try:
                    value = value()
                except Exception as exx:
                    self.ITEM.METHODS__EXX.update({name: exx})
                    continue

            # print(f"{name=}/{attr_obj=}/type={type(attr_obj)}/elementary={isinstance(attr_obj, TYPES.ELEMENTARY)}")

            # PLACE VALUE ---------------------------------------------------------------------
            if TypeAux(value).check__elementary_single():
                if attr_is_method:
                    self.ITEM.METHODS__ELEMENTARY_SINGLE.update({name: value})
                else:
                    self.ITEM.PROPERTIES__ELEMENTARY_SINGLE.update({name: value})

            elif TypeAux(value).check__elementary_collection():
                if attr_is_method:
                    self.ITEM.METHODS__ELEMENTARY_COLLECTION.update({name: value})
                else:
                    self.ITEM.PROPERTIES__ELEMENTARY_COLLECTION.update({name: value})

            else:
                if attr_is_method:
                    self.ITEM.METHODS__OBJECTS.update({name: value})
                else:
                    self.ITEM.PROPERTIES__OBJECTS.update({name: value})

    # =================================================================================================================
    def _print_line__group_separator(self, name: str) -> str:
        """
        GOAL MAIN - print!
        GOAL SECONDARY - return str - just for tests!!!
        """
        result = "-" * 10 + f"{name:-<90}"      # here is standard MAX_LINE_LEN
        print(result)
        return result

    def _print_line__name_type_value(self, name: Optional[str] = None, type_replace: Optional[str] = None, value: Union[None, Any, ItemInternal] = None, intend: Optional[int] = None) -> str:
        # -------------------------------
        name = name or ""
        if isinstance(value, ItemInternal):
            name = ""
        block_name = f"{name}"

        # -------------------------------
        block_type = f"{value.__class__.__name__}"
        if isinstance(value, ItemInternal):
            block_type = f"{value.KEY.__class__.__name__}:{value.VALUE.__class__.__name__}"
        if type_replace is not None:
            block_type = type_replace

        # -------------------------------
        intend = intend or 0
        if isinstance(value, ItemInternal):
            intend = 1

        _block_intend = "\t" * intend

        # -------------------------------
        block_value = f"{value}"
        if isinstance(value, ItemInternal):
            block_type = f"{value.KEY}:{value.VALUE}"
        elif TypeAux(value).check__exception():
            block_value = f"{value!r}"

        # -------------------------------
        result = f"{block_name:20}\t{block_type:12}:{_block_intend}{block_value}"

        if self.MAX_LINE_LEN and len(result) > self.MAX_LINE_LEN:
            result = result[:self.MAX_LINE_LEN - 3*2] + "..."

        # --------------------------------------------------------------------------------------
        print(result)

        # -------------------------------
        if name and str(value) != repr(value) and str(value) != str(block_value) and not TypeAux(value).check__exception():
            # additional print repr()
            self._print_line__name_type_value(name=None, type_replace="__repr()", value=repr(value))

        return result

    # =================================================================================================================
    def _print_block__head(self) -> None:
        # start printing ----------------------------------
        name = f"{self.__class__.__name__}.print"
        self._print_line__group_separator(name.upper())

        print(f"str(SOURCE)={str(self.SOURCE)}")
        print(f"repr(SOURCE)={repr(self.SOURCE)}")
        print(f"type(SOURCE)={type(self.SOURCE)}")

        try:
            mro = self.SOURCE.__class__.__mro__
        except:
            mro = self.SOURCE.__mro__

        mro = [cls.__name__ for cls in mro]

        print(f"mro(SOURCE)={mro}")

        # SETTINGS ----------------------------------------
        name = "SETTINGS"
        self._print_line__group_separator(name)

        print(f"{self.NAMES__USE_ONLY_PARTS=}")
        print(f"{self.NAMES__SKIP_FULL=}")
        print(f"{self.NAMES__SKIP_PARTS=}")

        print(f"{self.HIDE_BUILD_IN=}")
        print(f"{self.LOG_ITER=}")

        print(f"{self.MAX_LINE_LEN=}")
        print(f"{self.MAX_ITER_ITEMS=}")

    def _print_block__name_value(self, name, value) -> None:
        # ALWAYS ---------------------------------------------------------------------------------
        self._print_line__name_type_value(name=name, value=value)
        if len(str(value)) <= self.MAX_LINE_LEN:
            return

        # COLLECTION -----------------------------------------------------------------------------
        if TypeAux(value).check__elementary_collection():
            # start some pretty style -------------------------------------
            if not isinstance(value, dict):
                _index = 0
                for item in value:
                    _index += 1
                    if self.MAX_ITER_ITEMS and _index > self.MAX_ITER_ITEMS:
                        self._print_line__name_type_value(name=None, type_replace="", value="...", intend=1)
                        break
                    self._print_line__name_type_value(name=None, value=item, intend=1)

            elif isinstance(value, dict):
                _index = 0
                for item_key, item_value in value.items():
                    _index += 1
                    if self.MAX_ITER_ITEMS and _index > self.MAX_ITER_ITEMS:
                        self._print_line__name_type_value(name=None, type_replace="", value="...", intend=1)
                        break
                    self._print_line__name_type_value(name=None, value=ItemInternal(item_key, item_value))

        # SINGLE/EXX/OBJECTS ---------------------------------------------------------------------
        if any([
            TypeAux(value).check__elementary_single(),
            TypeAux(value).check__exception(),
            TypeAux(value).check__instance(),
        ]):
            pass    # DONT USE RETURN HERE OR ELIF IN NEXT LINE!!!

    # =================================================================================================================
    def print(self) -> None:
        """print all params from object
        if callable - try to call it!
        """
        WRAPPER_MAIN_LINE = "="*min(90, self.MAX_LINE_LEN)  # here we need use less then

        print(WRAPPER_MAIN_LINE)
        self._print_block__head()
        self._item_reload()

        for group_name, group_values in self.ITEM.__getstate__().items():
            self._print_line__group_separator(group_name)

            if TypeAux(group_values).check__elementary_collection_not_dict():
                for pos, name in enumerate(group_values, start=1):
                    print(f"{pos}:\t{name}")
            else:
                for name, value in group_values.items():
                    self._print_block__name_value(name, value)
        print(WRAPPER_MAIN_LINE)

    # =================================================================================================================
    def print_diffs(self) -> None:
        pass
        # TODO: FINISH!


# =====================================================================================================================



from typing import *


# =====================================================================================================================
class PROJECT:
    # AUTHOR -----------------------------------------------
    AUTHOR_NAME: str = "Andrei Starichenko"
    AUTHOR_EMAIL: str = "centroid@mail.ru"
    AUTHOR_HOMEPAGE: str = "https://github.com/centroid457/"
    AUTHOR_NICKNAME_GITHUB: str = "centroid457"

    # PROJECT ----------------------------------------------
    NAME_IMPORT: str = "cli_user"
    KEYWORDS: List[str] = [
        "cli", "cli user", "cli sender",
        "os terminal", "os terminal sender", "os terminal user",
    ]
    CLASSIFIERS_TOPICS_ADD: List[str] = [
        # "Topic :: Communications",
        # "Topic :: Communications :: Email",
    ]

    # README -----------------------------------------------
    # add DOUBLE SPACE at the end of all lines! for correct representation in MD-viewers
    DESCRIPTION_SHORT: str = "send commands into system terminal"
    DESCRIPTION_LONG: str = """
    Designed to send commands into OS terminal
        """
    FEATURES: List[str] = [
        # "feat1",
        # ["feat2", "block1", "block2"],

        "send commands into OS terminal",
        "check if cli commands are accessible (special utilities is installed)",
        "access to standard parts of result in a simple ready-to-use form (stdout/stderr/retcode/full state)",
        "use batch timeout for list",
        "till_first_true",
        "counter/counter_in_list",
    ]

    # HISTORY -----------------------------------------------
    VERSION: Tuple[int, int, int] = (0, 1, 3)
    TODO: List[str] = [
        "..."
    ]
    FIXME: List[str] = [
        "..."
    ]
    NEWS: List[str] = [
        "fix linux tests"
    ]

    # FINALIZE -----------------------------------------------
    NAME_INSTALL: str = NAME_IMPORT.replace("_", "-")


# =====================================================================================================================



from typing import *


# =====================================================================================================================
class PROJECT:
    # PROJECT ----------------------------------------------
    NAME_IMPORT: str = "object_info"
    KEYWORDS: list[str] = [
        "object info",
        "object attributes", "object properties", "object methods",
        "print attributes", "print properties", "print methods",
    ]

    # README -----------------------------------------------
    # add DOUBLE SPACE at the end of all lines! for correct representation in MD-viewers
    DESCRIPTION_SHORT: str = "print info about object (attributes+properties+methods results)"
    DESCRIPTION_LONG: str = """
Designed to print info about object (properties+methods results)

But why? if we can use debugger directly?  
Reason:  
1. to get and save standard text info,  
it useful to keep this info for future quick eye sight without exact condition like other OS or device/devlist/configuration 
2. in debugger we cant see result of methods!  
try to see for example information from platform module! it have only methods and no one in object tree in debugger!  
```python
import platform

obj = platform
print(platform.platform())
pass    # place debug point here
```  
3. Useful if you wish to see info from remote SOURCE if connecting directly over ssh for example  
    """
    FEATURES: list[str] = [
        # "feat1",
        # ["feat2", "block1", "block2"],

        "print all properties/methods results",
        "show exceptions on methods/properties",
        "skip names by full/part names and use only by partnames",
        "separated collections by groups",
    ]

    # HISTORY -----------------------------------------------
    VERSION: tuple[int, int, int] = (0, 2, 20)
    TODO: list[str] = [
        "add TIMEOUT (use start in thread!) for print! use timeout for GETATTR!!!",
        [
            "realise PRINT_DIFFS=CHANGE_state/COMPARE_objects (one from different states like thread before and after start)!",
            "this is about to save object STATE!",
            "add parameter show only diffs or show all",
            "add TESTS after this step!",
        ],
        "apply asyncio.run for coroutine?",
        "merge items Property/Meth? - cause it does not matter callable or not (just add type info block)",

        "add check__instance_of_user_class",
    ]


# =====================================================================================================================



import pathlib
import datetime
from base_aux.aux_text.m1_text_aux import TextAux
from base_aux.aux_types.m2_info import ObjectInfo


try:
    import git  # GITPYTHON # need try statement! if not installed git.exe raise Exx even if module was setup!!!
except:
    msg = f"[ERROR] git - is not setup in OS"
    print(msg)


# =====================================================================================================================
class Git:
    """
    GOAL
    ----
    get last commit short info instead of hard-version
    """
    # settings -------------
    PATH: pathlib.Path | str = None

    # aux -------------
    REPO: git.Repo = None   # real object/only existed

    # -----------------------------------------------------------------------------------------------------------------
    def __init__(self, path: pathlib.Path | str = None):
        if path:
            self.PATH = pathlib.Path(path)
        else:
            self.PATH = pathlib.Path.cwd()

        try:
            self.REPO = git.Repo("../..")   # FIXME: need auto find ROOT!!!
            self.REPO = git.Repo(self.PATH)
        except Exception as exx:
            print(f"git.Repo={exx!r}")
            print(f"возможно GIT не установлен")

    def check_ready(self) -> bool:
        """
        GOAL
        ----
        check if all ready to work
        - git setup
        - repo created
        """
        if self.REPO:
            return True
        else:
            return False

    def check_status(self) -> bool:
        """
        GOAL
        ----
        check if you work in validated repo! no changes from last commit
        """
        pass

    # -----------------------------------------------------------------------------------------------------------------
    @property
    def COMMITTER(self) -> str | None:
        """
        EXAMPLE
        -------
        ndrei Starichenko
        """
        if self.REPO:
            return self.REPO.head.object.committer

    @property
    def BRANCH(self) -> str | None:
        """
        EXAMPLE
        -------
        main
        """
        if self.REPO:
            try:
                result = self.REPO.active_branch.name
            except Exception as exx:
                msg = f"[GIT] DETACHED HEAD - you work not on last commit on brange! {exx!r}"
                print(msg)
                result = "*DETACHED_HEAD*"
            return result

    @property
    def SUMMARY(self) -> str | None:
        """
        actual commit text

        EXAMPLE
        -------
        [Text] add shortcut_nosub
        """
        if self.REPO:
            return self.REPO.commit().summary

    @property
    def HEXSHA(self) -> str | None:
        """
        NOTE
        ----
        more useful work with 8 chars! that's enough!

        EXAMPLE
        -------
        9fddeb5a9bed20895d56dd9871a69fd9dee5fbf7
        """
        if self.REPO:
            return self.REPO.head.object.hexsha

    @property
    def HEXSHA8(self) -> str | None:
        """
        derivative for main HEXSHA cut by 8 chars

        EXAMPLE
        -------
        9fddeb5a
        """
        if self.REPO:
            return self.HEXSHA[:8]

    @property
    def DATETIME(self) -> datetime.datetime | None:
        """
        EXAMPLE
        -------
        2024-12-05 11:30:17+03:00
        """
        if self.REPO:
            return self.REPO.head.object.committed_datetime

    # -----------------------------------------------------------------------------------------------------------------
    def info_string(self) -> str:
        """
        EXAMPLE
        -------
        git_mark='[git_mark//main/zero/Andrei Starichenko/ce5c3148/2024-12-04 18:39:10]'
        """
        if self.REPO:
            branch = TextAux(self.BRANCH).shortcut(15)
            summary = TextAux(self.SUMMARY).shortcut(15)
            dt = TextAux(self.DATETIME).shortcut_nosub(19)

            result = f"{branch}/{summary}/{self.COMMITTER}/{self.HEXSHA8}/{dt}"

        else:
            result = f"возможно GIT не установлен"

        git_mark = f"[git_mark//{result}]"
        print(f"{git_mark=}")
        return git_mark


# =====================================================================================================================
if __name__ == '__main__':
    from base_aux.aux_types import *
    ObjectInfo(Git()).print()


# =====================================================================================================================



from typing import *

from base_aux.aux_values.m0_novalue import NoValue
from base_aux.base_statics.m2_exceptions import *
from base_aux.aux_cmp_eq.m2_eq_aux import *


# =====================================================================================================================
TYPE__VARIANT = Union[str, Any]
TYPE__VARIANTS = list[TYPE__VARIANT] | NoValue


# =====================================================================================================================
# FOXME: DEPRECATE for new Eq style

class ValueVariants:
    """
    used to keep separated VALUE and measure unit

    SAME AS - ValueEqValid_Variants
    -------------------------------
    here is only validating and keep passed value
    in ValueVariants - final value used from exact Variants!

    GOAL
    ----
    1. get first associated value
    2. validate item by variants
    """
    # NOTE: NOTS DEPRECATE!!!
    # TODO: combine with ValueUnit - just add ACCEPTABLE(*VARIANTS) and rename UNIT just as SUFFIX!

    # SETTINGS -----------------------
    IGNORECASE: bool = True
    VARIANTS: TYPE__VARIANTS = NoValue
    VALUE_DEFAULT: Any = NoValue

    # DATA ---------------------------
    __value: Any = NoValue

    def __init__(self, value: Union[str, Any] = NoValue, variants: TYPE__VARIANTS = NoValue, ignorecase: bool = None):
        """
        """
        if ignorecase is not None:
            self.IGNORECASE = ignorecase

        self._variants_apply(variants)

        if value != NoValue:
            self.VALUE = value
            self.VALUE_DEFAULT = self.VALUE

        self._variants_apply(variants)  # need secondary!!!

    def _variants_apply(self, variants: set[Union[str, Any]] | NoValue = NoValue) -> None:
        if variants is not NoValue:
            self.VARIANTS = variants

        if self.VARIANTS is NoValue:
            if self.VALUE is not NoValue:
                self.VARIANTS = [self.VALUE, ]
            # else:
            #     self.VARIANTS = set()

    def __str__(self) -> str:
        return f"{self.VALUE}"

    def __repr__(self) -> str:
        """
        used as help
        """
        return f"{self.VALUE}{self.VARIANTS}"

    def __eq__(self, other):
        if isinstance(other, ValueVariants):
            if other.VALUE == NoValue:
                return self.VALUE in other
            else:
                other = other.VALUE

        if self.VALUE == NoValue:
            return self.value_validate(other)

        # todo: decide is it correct using comparing by str()??? by now i think it is good enough! but maybe add it as parameter
        if self.IGNORECASE:
            return (self.VALUE == other) or (str(self.VALUE).lower() == str(other).lower())
        else:
            return (self.VALUE == other) or (str(self.VALUE) == str(other))

    def __len__(self):
        return len(self.VARIANTS or [])

    def __iter__(self):
        yield from self.VARIANTS

    def __contains__(self, item) -> bool:
        """
        used to check compatibility
        """
        return self.value_validate(item)

    def __getitem__(self, item: int) -> Any:
        return self.VARIANTS[item]

    @property
    def VALUE(self) -> Any:
        return self.__value

    @VALUE.setter
    def VALUE(self, value: Any) -> Optional[NoReturn]:
        variant = self.value_get_variant(value)
        if variant != NoValue:
            self.__value = variant
        else:
            raise Exx__ValueNotValidated()

    def value_get_variant(self, value: Any) -> TYPE__VARIANT | NoValue:
        for variant in self.VARIANTS:
            if EqAux(variant).check_doubleside__bool(value):
                return variant

            if self.IGNORECASE:
                result = str(variant).lower() == str(value).lower()
            else:
                result = str(variant) == str(value)
            if result:
                return variant

        return NoValue

    def value_validate(self, value: Any) -> Any | None:
        return self.value_get_variant(value) != NoValue

    def reset(self) -> None:
        """
        set VALUE into default only if default is exists!
        """
        if self.VALUE_DEFAULT != NoValue:
            self.VALUE = self.VALUE_DEFAULT


# =====================================================================================================================



from typing import *
import numpy as np
import pandas as pd


# =====================================================================================================================
class Exx_TimeSeries(Exception):
    pass


# =====================================================================================================================
class HistoryShifted_Shrink:
    SOURCE: np.array = None
    SHIFT: int = None

    def __init__(self, source: np.array, shift: int = 1):
        self.SOURCE = source
        self.SHIFT = shift

        if shift < 1:
            msg = f"incorrect {shift=}"
            raise Exx_TimeSeries(msg)

    # FIELDS ----------------------------------------------------------------------------------------------------------
    def _get_fields(self) -> dict[str, Any]:
        """
        {
            'time': (dtype('int64'), 0),
            'open': (dtype('float64'), 8),
            'high': (dtype('float64'), 16),
            'low': (dtype('float64'), 24),
            'close': (dtype('float64'), 32),
            'tick_volume': (dtype('uint64'), 40),
            'spread': (dtype('int32'), 48),
            'real_volume': (dtype('uint64'), 52)
        }
        """
        # ['time', 'open', 'high', 'low', 'close', 'tick_volume', 'spread', 'real_volume']
        return self.SOURCE.dtype.fields

    # SHRINK ----------------------------------------------------------------------------------------------------------
    def shrink(self) -> np.array:
        if self.SHIFT == 1:
            return self.SOURCE
        windows = self._windows_get()
        result = self._windows_shrink(windows)
        return result

    # ------------------------------------------------------------------------------------------------------
    def _windows_get(self) -> np.array:
        bars_windows = np.lib.stride_tricks.sliding_window_view(x=self.SOURCE, window_shape=self.SHIFT)
        bars_windows_stepped = bars_windows[::self.SHIFT]
        return bars_windows_stepped

    def _windows_shrink(self, windows: np.array) -> np.array:
        result: Optional[np.array] = None
        for window in windows:
            void_new = self._window_shrink(window)
            try:
                result = np.concatenate([result, [void_new]])
            except Exception as exx:
                # if no elements
                # print(f"{exx!r}")
                result = np.array([void_new])
        return result

    def _window_shrink(self, window: np.array) -> np.void:
        void_new = window[0].copy()

        void_new["time"] = window["time"].max()
        void_new["open"] = window["open"][-1]
        void_new["high"] = window["high"].max()
        void_new["low"] = window["low"].min()
        void_new["close"] = window["close"][0]
        void_new["tick_volume"] = window["tick_volume"].sum()    # may be incorrect
        void_new["spread"] = void_new["high"] - void_new["low"]    # may be incorrect
        void_new["real_volume"] = window["real_volume"].sum()

        return void_new


# =====================================================================================================================
class HistoryShifted_Simple:
    """History manager when important only one column in calculations!
    such as RSI/WMA tipically use only close values from timeSeries!
    """
    SOURCE: np.array = None
    COLUMN: str = "close"
    SHIFT: int = None

    def __init__(self, source: np.array, shift: int = 1, column: Optional[str] = None):
        self.SOURCE = source
        self.COLUMN = column
        self.SHIFT = shift

        if shift < 1:
            msg = f"incorrect {shift=}"
            raise Exx_TimeSeries(msg)


# =====================================================================================================================



from base_aux.base_singletons.m1_singleton import *

class MySingleton(SingletonCallMeta):
    pass

class MySingleton2(SingletonCallMeta):
    pass

class MySingleton(metaclass=SingletonCallMetaType):
    pass


# ===============================
# 2. access to created instances
from base_aux.base_singletons.m1_singleton import *

class Victim1(SingletonCallMeta):
    attr = 1

class Victim2(SingletonCallMeta):
    attr = 2

assert SingletonCallMeta._SINGLETONS == []
Victim1()
assert SingletonCallMeta._SINGLETONS == [Victim1(), ]
assert Victim1._SINGLETONS == [Victim1(), ]
assert Victim1()._SINGLETONS == [Victim1(), ]
Victim2()
assert SingletonCallMeta._SINGLETONS == [Victim1(), Victim2(), ]


# ===============================
# 3. NOTICE: all your Singletons must be only last classes!
# don't use nesting from any Your Singletons!
from base_aux.base_singletons import *

class MySingleton(SingletonCallMeta):  # OK
    pass

class MySingleton2(MySingleton):  # WRONG
    pass


import time

from base_aux.testplans.tc import *
from base_aux.valid.m2_valid_base2_derivatives import *
from base_aux.valid.m3_valid_chains import *


# =====================================================================================================================
class TestCase(TestCaseBase):
    ASYNC = True
    DESCRIPTION = "direct1"

    def run__wrapped(self):
        time.sleep(0.1)
        self.details_update({"detail_value": self.DEVICES__BREEDER_INST.DUT.VALUE})
        result_chain = ValidChains(
            [
                Valid(value_link=True, name="TRUE"),
            ],
        )
        return result_chain


# =====================================================================================================================



from typing import *
from enum import Enum
from PyQt5.QtCore import QThread
import time

from base_aux.servers.m0_url import UrlCreator

from fastapi import FastAPI, Path, Query, Body, Response
from fastapi.responses import JSONResponse, RedirectResponse

from pydantic import BaseModel
import uvicorn

from starlette import status
from starlette.responses import Response

from base_aux.loggers.m1_logger import Logger


# =====================================================================================================================
class DataExample:
    int = 1
    float = 1.2
    str = "string"
    list = [1, "2"]
    dict = {1: "2", "22": 11}       # by FastAPI return={"1":"2","22":11}


# =====================================================================================================================
def _minimal():
    import uvicorn
    from fastapi import FastAPI, Response
    from fastapi.responses import RedirectResponse

    app = FastAPI()

    @app.get("/")
    async def redirect() -> Response:
        return RedirectResponse(url="/docs")

    @app.get("/{path}")
    async def hello(path):
        print(path)
        return path

    uvicorn.run(app, host="localhost", port=80)


# =====================================================================================================================
def create_app__FastApi(self: Any = None, data: Any = None) -> FastAPI:
    class Item(BaseModel):
        value: int

    # UNIVERSAL ======================================================
    if data is None:
        data = DataExample()

    app = FastAPI()
    app.data = data
    app.LOGGER = Logger("FastAPI").LOGGER

    # WORK -----------------------------------------------------------
    @app.get("/")
    async def redirect() -> Response:
        return RedirectResponse(url="/docs")

    @app.post("/post/dict")
    async def post(item: Item):
        app.LOGGER.debug(item)
        data.dict.update(item)
        print(data.dict)
        return item

    pass
    pass
    pass
    pass
    pass
    pass

    # EXAMPLES =======================================================
    pass

    # NOT FOUND ------------------------------------------------------
    async def NOT_FOUND():
        pass
        # by default if not specified on any resuets return - {"detail": "Not Found"}

    # ORDER ----------------------------------------------------------
    @app.get("/order")
    async def order_1():
        return "ALWAYS WORK FIRST MATCH!!!"

    @app.get("/order")
    async def order_2():
        return "this will work NEVER!!!"

    # CASE_SENSE -----------------------------------------------------
    @app.get("/CASE")
    async def case_sense():
        return "CaseSensitive!!!"

    # SLASHES ---------------------------------------------------------
    # COUNT -----------------
    @app.get("/slash_finish")   # this is another path! '/slash/' and '/slash' - is different!!!
    async def slash_single():
        return {"slash_finish": ""}

    @app.get("/slash_finish/")
    async def slash_single():
        return {"slash_finish": "/"}

    @app.get("/slash_finish//")
    async def slash_double():
        return {"slash_finish": "//"}

    @app.get("/slash_finish///")
    async def slash_triple():
        return {"slash_finish": "///"}

    # START -----------------
    @app.get("/slash")
    async def slash_start():
        return "/slash"

    @app.get("slash")       # never met! so you need use at least one start slash!!!
    async def slash_nostart():
        return "slash"

    @app.get("slash_no")       # never met! even by 'http*.../slash_no'
    async def slash_no():
        return "slash_no"

    # FUNC_NAMES ------------------------------------------------------
    @app.get("/1")
    async def same_name():
        return {"same_name": 1}

    @app.get("/2")
    async def same_name():    # decorator is the unic! name for func is not important!!!
        return {"same_name": 2}

    # TYPES -----------------------------------------------------------
    @app.get("/return_types/str")
    async def return_types():
        return "str"

    @app.get("/return_types/int")
    async def return_types():
        return 123

    @app.get("/return_types/float")
    async def return_types():
        return 111.222

    @app.get("/return_types/true")
    async def return_types():
        return True         # true

    @app.get("/return_types/false")
    async def return_types():
        return False        # false

    @app.get("/return_types/none")
    async def return_types():
        return None         # null

    @app.get("/return_types/list")
    async def return_types():
        return [1,"2",]

    @app.get("/return_types/dict")
    async def return_types():
        return {"key": "value"}

    @app.get("/return_types/model")
    async def return_types():
        """
        You can also return Pydantic models (you'll see more about that later).
        There are many other aux_types and models that will be automatically converted to JSON (including ORMs, etc). Try using your favorite ones, it's highly probable that they are already supported.
        """
        return "FINISH!!!!"     # TODO: FIXME: FINISH!!!

    # PARAMS ----------------------------------------------------------------------------------------------------------
    # VALIDATION ------------------------
    @app.get("/validation/")
    async def read_items(q: Annotated[str | None, Query(max_length=50)] = None):
        results = {"items": [{"item_id": "Foo"}, {"item_id": "Bar"}]}
        if q:
            results.update({"q": q})
        return results

    # PATH_PARAMS -----------------------------------------------------------
    # chars ----------------------
    @app.get("/path_chars/{name}")
    async def path(name: str):
        """
            http://localhost:8000/path_chars/123 -> "123"

            http://localhost:8000/path_chars/111-222!@$%5E*()_-+=.,%3C%3E%7B%7D[]%60~#CommentFinal  -> "111-222!@$^*()_-+=.,<>{}[]`~"

        COMMENTS = only FINAL!
            http://localhost:8000/path_chars/111-222#CommentFinal/hello     -> "111-222"
        """
        return name

    # order_matters ----------------------
    @app.get("/path_param/fixed")
    async def order_matters():
        return {"path_param": "FIXED PATH PLACE FIRST!"}

    @app.get("/path_param/{as_param}")
    async def order_matters(as_param: str):
        return {"path_param": as_param}

    # UNIVERSAL+NONE_AS_DEFAULT-----------------------------
    @app.get("/path_param_try_def/{name}/{value}")  # path item must used! if not - 404=NOT FOUND
    # def path_params(name, value=None):  # so never use None! -  it would not affect anything!
    async def path_params(name, value):
        """
            http://localhost:8000/path_param_try_def/111/       ->{"detail":"Not Found"}
            http://localhost:8000/path_param_try_def/111/222    ->{"name":"111","value":"222"}
        """
        return {"name": name, "value": value}

    @app.get("/path_params__validate_type/{name}/{value}")
    async def path_params__validate_type(name: str, value: int):
        """
        VALIDATION will try to change type
        so
            /1 -> int -> 1
            /1 -> str -> "1"
        for error
            /wrong -> int -> {"detail":[{"type":"int_parsing","loc":["path","value"],"msg":"Input should be a valid integer, unable to parse string as an integer","input":"wrong","url":"https://errors.pydantic.dev/2.6/v/int_parsing"}]}

        http://localhost:8000/params_validate/hello/1 -> {"name":"hello","value":1}
        """
        return {"name": name, "value": value}

    @app.get("/params_validate/{name}/{value2}")
    async def params_validate__overload(name: str, value2: str):
        """
        OVERLOAD WOULDNOT WORK!!!
        INFO:     127.0.0.1:55128 - "GET /params_validate/hello/wrong HTTP/1.1" 422 Unprocessable Entity
        """
        return {"name": name, "value2": value2}

    # PATH_LONG ----------------------
    @app.get("/files/{file_path:path}")
    async def path_long(file_path: str):
        """
            http://localhost:8000/files/hello/123 -> {"file_path":"hello/123"}
            http://localhost:8000/files//hello/123 -> {"file_path":"/hello/123"}
        """
        return {"file_path": file_path}

    # ENUM ----------------------
    class Model_Enum(str, Enum):    # important define by both STR+ENUM!!!
        alexnet = "alexnet"
        resnet = "resnet"
        lenet = "lenet"

    @app.get("/path_params__validate_enum/{name_by_enum}")
    async def path_params__validate_enum(name_by_enum: Model_Enum):
        if name_by_enum is Model_Enum.alexnet:
            return {"model_name": name_by_enum, "message": "Deep Learning FTW!"}

        if name_by_enum.value == "lenet":
            return {"model_name": name_by_enum, "message": "LeCNN all the images"}

        return {"model_name": name_by_enum, "message": "Have some residuals"}

    # QUERY_PARAMS ------------------------------------------------------------------
    """
    OPTIONAL CORRECT SYNTAX
    ALL YOU NEED IS SET NONE AS DEFAULT VALUE FOR ANY TYPE!!! - Having a default value of any type, including None, makes the parameter optional (not required).
    
    
    FastAPI will know that the value of q is not required because of the default value = None. - SEEMS IT BEST! JUST USE "= NONE"!!!
    The Union in Union[str, None] is not used by FastAPI, but will allow your editor to give you better support and detect errors. - JUST FOR IDE DEBUGGING
        
    Union[str, None] (or str | None in Python 3.10)
        q: str | None = None    - AS BEST IN DOCS!!! for new python!
        q: Union[str, None] = None  - for old python!
    """

    @app.get("/query_chars")
    async def query(q: Union[str, None] = None):
        """
        QUERY params are all that not parametrizes
        The query is the set of key-value pairs that go after the ? in a URL, separated by & characters.

        FOR BLANC - RETURN NULL
            http://localhost:8000/query?        -> {"q":null}
            http://localhost:8000/query?123     -> {"q":null}

            http://localhost:8000/query?q       -> {"q":""}
            http://localhost:8000/query?q=      -> {"q":""}
            http://localhost:8000/query?q=123   -> {"q":"123"}

            http://localhost:8000/query/?q=     -> {"q":""}    # query is not depends on last finishing slash!!!
            http://localhost:8000/query//?q=    -> {"q":""}    # query is not depends on last finishing slash!!!

            http://localhost:8000/query?hello=123 -> {"q":null}     # use not defined query param! - no error!

        try send None
            http://localhost:8000/query?        -> {"q":null}   # seems only one way is not send any!!!
            http://localhost:8000/query?q=null  -> {"q":"null"}
            http://localhost:8000/query?q=none  -> {"q":"none"}
            http://localhost:8000/query?q=None  -> {"q":"None"}

        quotation marks
            http://localhost:8000/query?q="none"    -> {"q":"\"none\""}
            http://localhost:8000/query?q=%22none%22 - past to here by copy from address in brouser!

        simbols
            http://localhost:8000/query?q=\ /_- =[]{}()<>:;,.*!?@$%^~#CommentFinal -> {"q":"\\ /_- =[]{}()<>:;,.*!?@$%^~"}
        """
        return {"q": q}

    @app.get("/query_several")
    async def query(q1: Union[str, None] = None, q2: Union[str, None] = None):
        """
            http://localhost:8000/query_several?q2=222&q1=111 -> {"q1":"111","q2":"222"}

        comment - only final!
            http://localhost:8000/query_several?q1=123#CommentFinal&q2=willNotUsed -> {"q1":"123","q2":null}

        """
        return {"q1": q1, "q2": q2}

    @app.get("/query_list")
    async def query(
            q: Annotated[Union[list[str], None], Query()] = None,
            q2: Annotated[list[str], Query()] = ["foo", "bar"],
    ):
        """
        To declare a query parameter with a type of list, you need to explicitly use Query, otherwise it would be interpreted as a request body.
            http://localhost:8000/query_list?q=1&q=2    -> {"q":["1","2"]}
            http://localhost:8000/query_list?q=1        -> {"q":["1"]}
            http://localhost:8000/query_list            -> {"q":null}

            http://localhost:8000/query_list            -> {"q":null,"q2":["foo","bar"]}
        """
        return {"q": q, "q2": q2}

    @app.get("/query_with_path/{param}")
    async def query(param: int, q: Union[str, None] = None):
        return {"param": param, "q": q}

    @app.get("/query_validate")
    async def query(q: Union[bool, None] = None, qenum: Model_Enum = None):
        """
        BOOL - any extended convertation!
            http://localhost:8000/query_validate            -> {"q":null}

            http://localhost:8000/query_validate?q=1        -> {"q":true}

            http://localhost:8000/query_validate?q=TRUE     -> {"q":true}
            http://localhost:8000/query_validate?q=True     -> {"q":true}
            http://localhost:8000/query_validate?q=true     -> {"q":true}

            http://localhost:8000/query_validate?q=Y        -> {"q":true}
            http://localhost:8000/query_validate?q=Yes      -> {"q":true}

            http://localhost:8000/query_validate?q=ON      -> {"q":true}

            http://localhost:8000/query_validate?q=0        -> {"q":false}

        BOOL WRONG
            http://localhost:8000/query_validate?q=         -> {"detail":[{"type":"bool_parsing","loc":["query","q"],"msg":"Input should be a valid boolean, ....
            http://localhost:8000/query_validate?q="yes"
        """
        return {"q": q}

    @app.get("/query_required")
    async def query(q: str):
        """
        when not passed
            http://localhost:8000/query_required
                -> {"detail":[{"type":"missing","loc":["query","q"],"msg":"Field required","input":null,"url":"https://errors.pydantic.dev/2.6/v/missing"}]}

        when passed
            http://localhost:8000/query_required?q=111  -> {"q":"111"}
        """
        return {"q": q}

    @app.get("/query_optional")
    async def query(q: str = None):
        """
        its enough to set None + not need notice in Annotation!

            http://localhost:8000/query_optional    -> {"q":null}
        """
        return {"q": q}

    # POST ------------------------------------------------------------------------------------------------------------
    """
    To send data, you should use one of: POST (the more common), PUT, DELETE or PATCH.
    Sending a body with a GET request has an undefined behavior in the specifications, nevertheless, it is supported by FastAPI, only for very complex/extreme use cases.
    As it is discouraged, the interactive docs with Swagger UI won't show the documentation for the body when using GET, and proxies in the middle might not support it.
    """

    # MODELS ---------------------------------
    class Item(BaseModel):
        name: str
        description: str | None = None
        price: float
        tax: float | None = None

    class User(BaseModel):
        username: str
        full_name: str | None = None

    @app.post("/post/nobody/start")
    async def post():
        return

    @app.post("/post/body/single/")
    async def post(item: Item | None = None):
        """
        BODY could be Optional!!!

        Expected BODY simple
            {
                "name": "Foo",
                "description": "The pretender",
                "price": 42.0,
                # "tax": 3.2
            }
        """
        # return item

        item_dict = item.dict()
        if item.tax:
            price_with_tax = item.price + item.tax
            item_dict.update({"price_with_tax": price_with_tax})

        if item:
            item_dict.update({"item": item})

        return item_dict

    @app.post("/post/body/multy/")
    async def post(item: Item | None = None, user: User | None = None, extra_singular: Annotated[int | None, Body()] = None):
        """
        Expected BODY as complicated
            {
                "item": {
                    "name": "Foo",
                    "description": "The pretender",
                    "price": 42.0,
                    "tax": 3.2
                },
                "user": {
                    "username": "dave",
                    "full_name": "Dave Grohl"
                },
                "extra_singular": None,
            }
        """
        results = {"item": item, "user": user, "extra_singular": extra_singular}
        return results

    @app.post("/post/all_params/{item_id}")
    async def post(item_id: int, item: Item, q: str | None = None):
        """
        RECOGNITION ORDER!
        The function parameters will be recognized as follows:
            If the parameter is also declared in the path, it will be used as a path parameter.
            If the parameter is of a singular type (like int, float, str, bool, etc) it will be interpreted as a query parameter.
            If the parameter is declared to be of the type of a Pydantic model, it will be interpreted as a request body.
        """
        return {"item_id": item_id, **item.dict(), "q": q}

    # RESPONSE ----------------------------------------------------------------
    @app.get("/redirect")
    async def redirect() -> Response:
        return RedirectResponse(url="https://ya.ru")    # INFO: 127.0.0.1:61637 - "GET /redirect HTTP/1.1" 307 Temporary Redirect

    @app.get("/")
    async def redirect() -> Response:
        return RedirectResponse(url="/docs")

    @app.get("/json")
    async def json() -> Response:
        return JSONResponse(
            content={"message": "Here's your interdimensional portal."},
            status_code=200,
        )

    # RESULT ------------------------------------------------------------------
    return app


# =====================================================================================================================
def create_app__APIRouter() -> 'APIRouter':
    pass

    # from fastapi import APIRouter
    # from starlette import status
    # from starlette.responses import Response

    # from bot import proceed_release
    # from models import Body, Actions
    #
    # api_router = APIRouter()  # noqa: pylint=invalid-name
    #
    # @api_router.post("/release/")
    # async def release(*,
    #                   body: Body,
    #                   chat_id: str = None,
    #                   release_only: bool = False):
    #
    #     if (body.release.draft and not release_only) or body.action == Actions.released:
    #         res = await proceed_release(body, chat_id)
    #         return Response(status_code=res.status_code)
    #     return Response(status_code=status.HTTP_200_OK)
    #
    # return api_router


# =====================================================================================================================
pass
pass
pass
pass


def start_1__by_terminal(app: FastAPI) -> None:
    """
    uvicorn main:app
    uvicorn main:app --reload

    2=FROM PYTHON CODE
    https://stackoverflow.com/questions/62856818/how-can-i-run-the-fast-api-server-using-pycharm
    """
    pass


# =====================================================================================================================
class ServerFastApi_Thread(Logger, QThread):
    """
    WORK IN both LINUX/Win!!!
    """
    PORT: int = 80
    HOST: str = "0.0.0.0"
    # HOST: str = "localhost"
    """
    HOST SETTINGS RULES
    localhost - CANT ACCESS BY HOST_IP! only
        http://localhost/ - OK!
        http://127.0.0.1/ - OK!
        http://192.168.75.140/ - FAIL!!!
    
    0.0.0.0 - ALL ARE OK!!!
        http://localhost/ - OK!
        http://127.0.0.1/ - OK!
        http://192.168.75.140/ - OK!!!
    """

    data: Any = None
    create_app: Callable[[Any], FastAPI] = create_app__FastApi

    @property
    def ROOT(self) -> str:
        return UrlCreator().URL_create(host=self.HOST, port=self.PORT)

    def __init__(self, app: FastAPI = None, data: Any = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if data is None:
            data = self.data

        if app is None:
            app = self.create_app(data=data)
        self.app = app
        self.data = app.data

    def run(self):
        self.LOGGER.debug("run")
        uvicorn.run(self.app, host=self.HOST, port=self.PORT)

    def start(self, *args, **kwargs):
        super().start()
        time.sleep(1)


def start_2__by_thread(app: FastAPI = None) -> Never:
    server = ServerFastApi_Thread(app)
    # server.run()
    server.start()
    server.wait()


# =====================================================================================================================
def start_3__by_asyncio(app: FastAPI) -> Never:
    pass

    # async uvicorn.run(app, host="0.0.0.0", port=8000)


# =====================================================================================================================
# class ServerFastApi__Wrong1(FastAPI, QThread):
#     HOST = "0.0.0.0"
#     PORT = 8000
#
#     DATA = DataExample()
#
#     @get("/data/{attr}")    # EXX
#     def hello(self, attr):
#         return getattr(self.DATA, attr)
#
#     def run(self):
#         uvicorn.run(self, host=self.HOST, port=self.PORT)


# class ServerFastApi__Wrong2(FastAPI, QThread):
#     HOST = "0.0.0.0"
#     PORT = 8000
#
#     DATA = DataExample()
#
#     app = FastAPI()
#
#     def __init__(self, *args, **kwargs):
#         self.add_routs()
#         super().__init__(*args, **kwargs)
#
#     @classmethod
#     def add_routs(cls):               # will not create pathes
#         @cls.app.get("/data/{attr}")
#         async def hello(attr):
#             return getattr(cls.DATA, attr)
#
#     def run(self):
#         uvicorn.run(self, host=self.HOST, port=self.PORT)


# =====================================================================================================================
if __name__ == "__main__":
    start_2__by_thread()


# =====================================================================================================================



from base_aux.base_statics.m2_exceptions import *
from base_aux.aux_types.m1_type_aux import TypeAux
from base_aux.aux_callable.m1_callable_aux import *

from base_aux.aux_attr.m1_attr1_aux import AttrAux


# =====================================================================================================================
# FIXME: rebuild to separated special class RaiseIfPositive
#   by now it seems to sophisticated!


# =====================================================================================================================
class GetattrPrefixInst:
    """
    this is just a Base!

    EXAMPLES
    --------
    see GetattrPrefixInst_RaiseIf with tests
    """
    GETATTR_PREFIXES: list[str] = []

    def __getattr__(self, item: str) -> Any | Callable | NoReturn:
        """
        SHARING PARAMS BETWEEN CALLABLE PREFIX/ITEM
        -------------------------------------------
        you can not use params! and not use callables (static attributes are available)!
        args - all - goes for ITEM
        kwargs - all uppercase - goes for PREFIX (after changing by lower())
        kwargs - others - goes for ITEM directly (without changing case)!

        if you provide not existed args/kwargs - you will get direct corresponding exception like "print(hello=1) -> TypeError: 'hello' is an invalid keyword argument for print() "

        NOTE
        ----
        0. CaseInSensitive!
        1. prefix - callable with first parameter as catching item_value (may be callable or not).
        2.You always need to CALL prefixed result! even if you access to not callable attribute!
        """
        # print("-"*10)
        # print(f"{item=}start")
        # pretend DIRECT anycase name/prefix ----------
        item_original = AttrAux(self).anycase__name_original(item)
        if item_original:
            if item_original.lower() == item.lower():
                return getattr(self, item_original)

        # pretend PREFIX ----------
        for prefix in self.GETATTR_PREFIXES:
            # print(f"{prefix=}start")
            prefix_original = AttrAux(self).anycase__name_original(prefix)
            if not prefix_original:
                continue

            prefix_meth = getattr(self, prefix_original)

            # direct prefix ----------
            # if item.lower() == prefix.lower():
            #     return lambda *prefix_args, **prefix_kwargs: Lambda(prefix_meth).resolve_raise(*prefix_args, **prefix_kwargs)

            # prefix ----------
            if item.lower().startswith(prefix.lower()):
                item_name = item[len(prefix):]
                item_value = AttrAux(self).anycase__getattr(item_name)

                return lambda *meth_args, **meth_kwargs: CallableAux(prefix_meth).resolve_raise(
                    *[CallableAux(item_value).resolve_raise(*meth_args, **{k:v for k,v in meth_kwargs.items() if not k.isupper()}), ],
                    **{k.lower():v for k,v in meth_kwargs.items() if k.isupper()}
                )

        # print(3)
        raise AttributeError(item)


# =====================================================================================================================
class GetattrPrefixInst_RaiseIf(GetattrPrefixInst):
    """
    RULES
    -----
    """
    GETATTR_PREFIXES = ["raise_if__", "raise_if_not__"]     # dont use name like _*__*!

    # -----------------------------------------------------------------------------------------------------------------
    def raise_if__(self, source: Any, _reverse: bool | None = None, comment: str = "") -> None | NoReturn:
        result = CallableAux(source).resolve_exx()
        if TypeAux(result).check__exception() or bool(result) != bool(_reverse):
            raise Exx__GetattrPrefix_RaiseIf(f"[raise_if__/{_reverse=}]met conditions ({source=}/{comment=})")

    def raise_if_not__(self, source: Any, comment: str = "") -> None | NoReturn:
        return self.raise_if__(source=source, _reverse=True, comment=comment)


# =====================================================================================================================
def _example():
    class GetattrPrefixInst_RaiseIf_data(GetattrPrefixInst_RaiseIf):
        ATTR0 = 0
        ATTR1 = 1

        TRUE = True
        FALSE = False
        NONE = None

        def meth(self, value: Any = None):
            return value

        def METH(self, arg1=1, arg2=2):
            return arg1 == arg2

    assert GetattrPrefixInst_RaiseIf_data().TRUE is True
    assert GetattrPrefixInst_RaiseIf_data().true is True
    assert GetattrPrefixInst_RaiseIf_data().raise_if__none() is None

    # ON ATTRIBUTE --------------------
    # apply prefix+item
    GetattrPrefixInst_RaiseIf_data().raise_if__attr0()  # OK
    GetattrPrefixInst_RaiseIf_data().raise_if__attr1()  # Exx__GetattrPrefix_RaiseIf

    # ON METH -------------------------
    # args - one
    GetattrPrefixInst_RaiseIf_data().raise_if__meth(0)  # OK
    GetattrPrefixInst_RaiseIf_data().raise_if__meth(1)  # OK
    GetattrPrefixInst_RaiseIf_data().raise_if__meth(2)  # Exx__GetattrPrefix_RaiseIf

    # args - several
    GetattrPrefixInst_RaiseIf_data().raise_if__meth(1,2)  # OK
    GetattrPrefixInst_RaiseIf_data().raise_if__meth(2,2)  # Exx__GetattrPrefix_RaiseIf

    # kwargs
    GetattrPrefixInst_RaiseIf_data().raise_if__meth(1, arg2=2)  # OK
    GetattrPrefixInst_RaiseIf_data().raise_if__meth(2, arg2=2)  # Exx__GetattrPrefix_RaiseIf

    # send kwarg for PREFIX ----------
    GetattrPrefixInst_RaiseIf_data().raise_if__meth(1, arg2=2, comment="WRONG!")    # RAISE comment argument invalid for meth()!!!
    GetattrPrefixInst_RaiseIf_data().raise_if__meth(1, arg2=2, COMMENT="CORRECT!")  # OK
    GetattrPrefixInst_RaiseIf_data().raise_if__meth(2, arg2=2, COMMENT="CORRECT!")  # Exx__GetattrPrefix_RaiseIf


# =====================================================================================================================
if __name__ == "__main__":
    _example()


# =====================================================================================================================



from typing import *

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *


# =====================================================================================================================
# original names is not enough soft and comfortable for eyes!
# try to use ROUND NUMBERS!!! start use 0/100/200/255!!!
class COLOR_TUPLE_RGB:
    LIGHT_RED: tuple[int, int, int] = (255, 200, 200)
    LIGHT_GREEN: tuple[int, int, int] = (200, 255, 200)
    LIGHT_BLUE: tuple[int, int, int] = (200, 200, 255)
    LIGHT_YELLOW: tuple[int, int, int] = (255, 255, 100)
    LIGHT_GREY_240: tuple[int, int, int] = (240, 240, 240)   # #f2f2f2
    LIGHT_GREY_220: tuple[int, int, int] = (220, 220, 220)
    LIGHT_GREY_150: tuple[int, int, int] = (150, 150, 150)
    WHITE: tuple[int, int, int] = (255, 255, 255)


class MARGINS:
    _0000: QMargins = QMargins(0, 0, 0, 0)
    _5555: QMargins = QMargins(5, 5, 5, 5)
    _9595: QMargins = QMargins(9, 5, 9, 5)


class ALIGNMENT:
    T: Qt.Alignment = Qt.Alignment(Qt.AlignTop)
    TL: Qt.Alignment = Qt.Alignment(Qt.AlignTop | Qt.AlignLeft)
    TR: Qt.Alignment = Qt.Alignment(Qt.AlignTop | Qt.AlignRight)

    C: Qt.Alignment = Qt.Alignment(Qt.AlignVCenter | Qt.AlignHCenter)
    CH: Qt.Alignment = Qt.Alignment(Qt.AlignHCenter)
    CV: Qt.Alignment = Qt.Alignment(Qt.AlignVCenter)

    CT: Qt.Alignment = Qt.Alignment(Qt.AlignHCenter | Qt.AlignTop)
    CB: Qt.Alignment = Qt.Alignment(Qt.AlignHCenter | Qt.AlignBottom)
    CL: Qt.Alignment = Qt.Alignment(Qt.AlignVCenter | Qt.AlignLeft)
    CR: Qt.Alignment = Qt.Alignment(Qt.AlignVCenter | Qt.AlignRight)


# =====================================================================================================================



from typing import *

from base_aux.base_statics.m2_exceptions import *


# =====================================================================================================================
# TODO: need decision about flags order!!! NEED HUMAN ALWAYS except hex!!!
"""
work with bits (bitfields/flags/bites...)
designed for common work with bitfields-like aux_types
field_bytearray
flags/flags manipulate
bits/bit user/bit work/bits manipulate
"""


# =====================================================================================================================
class Bitfield:
    """
    IMPORTANT: ORDER for any field is NO HUMAN!!! or not???
    YOU NEED TO SEE ALWAYS AS REVERSE IN ANY REPRESENTATION!!!
    but when you indexind by obj[] used human order!!!
    """
    field_size: int = None
    _field_bytearray: bytearray = None

    def __init__(self, field_size: int):
        if field_size < 1:
            raise Exx__Incompatible(f"{field_size=}")

        self.field_size = field_size
        self._field_bytearray = bytearray((field_size + 7) // 8)

    # todo: clear_all+exact+several/setup_all+exact+several/ [by_pos/by_fields]
    # todo: listActiveFlags/list+iter [by_pos/by_fields]

    @classmethod
    def create_from_int(cls, flags: int, field_size: Optional[int] = None) -> 'Bitfield':
        raise NotImplemented
    # todo: FINISH!
    # todo: FINISH!
    # todo: FINISH!
    # todo: FINISH!
    # todo: FINISH!
    # todo: FINISH!
    # todo: FINISH!
    # todo: FINISH!

    def size_get_active(self) -> int:
        """get position of major valued flag
        """
        result = 0
        for pos, value in enumerate(self.list_bits()[::-1], start=1):
            if value == 1:
                result = pos
        return result

    @property
    def field_str(self) -> str:
        """in NOhuman order"""
        return "".join(map(str, self.list_bits()))

    def __getitem__(self, idx: int) -> Union[int, NoReturn]:
        if idx < -self.field_size or idx > self.field_size - 1:
            raise Exx__OutOfRange
        return self._field_bytearray[idx // 8] >> (idx % 8) & 1

    def __setitem__(self, idx: int, value: Union[int, bool, Any]) -> Optional[NoReturn]:
        mask = 1 << (idx % 8)
        if bool(value):
            self._field_bytearray[idx // 8] |= mask
        else:
            self._field_bytearray[idx // 8] &= ~mask

    def __len__(self) -> int:
        return self.field_size

    def count_activated(self) -> int:
        return bin(self.int()).count('1')

    # LISTS ---------------------------------------------------------------------------
    def list_bits(self) -> list[int]:
        """in NOhuman order
        """
        result = []
        for index in range(self.field_size):
            result.append(self[index])
        result.reverse()
        return result

    def list_activated_indexes(self) -> list[int]:
        """Return activated flag indexes in field!
        in human order (but order is not important!)

        in case of [101] -> [0, 2, ]
        """
        result = []
        for index in range(self.field_size):
            if self[index]:
                result.append(index)
        return result

    def list_activated_values(self) -> list[int]:
        """Return activated flag indexes in field!
        in human order (but order is not important!)

        in case of [101] -> [1, 4, ]
        """
        result = []
        for index in range(self.field_size):
            if self[index]:
                result.append(1 << index)
        return result

    # REPRESENTATIONS ---------------------------------------------------------------
    def __str__(self) -> str:
        return f"field[{self.field_str}]"

    def __repr__(self) -> str:
        return f"field[{self.field_str}]"

    def int(self) -> int:
        """
        print(int(bytearray(2).hex(), 16))
        """
        return int(self._field_bytearray.hex(), 16)

    def _bin_str(self) -> str:
        """common bin-format (without type prefix)

        print(bin(int(bytearray(2).hex())))     # 0b0
        """
        result = bin(self.int())[2:]
        result = "0" * (self.field_size - len(result)) + result
        return result

    def _hex_str(self) -> str:
        """common hex(full bytes)-format (without type prefix)
        print(bytearray(2).hex())
        """
        return self._field_bytearray.hex()


# =====================================================================================================================
class Flags(Bitfield):
    pass


# =====================================================================================================================
class Bits(Bitfield):
    pass


# =====================================================================================================================



import os
import re
from typing import Any, NoReturn

from base_aux.aux_attr.m2_annot1_aux import AnnotsAux
from base_aux.aux_attr.m4_kits import AttrKit_Blank
from base_aux.aux_types.m1_type_aux import TypeAux
from base_aux.base_inits.m3_nest_init_annots_attrs_by_kwargs import NestInit_AnnotsAttrByKwArgs
from base_aux.base_resolver.m1_resolver import NestCall_Resolve


class PvLoaderEnv(NestCall_Resolve):
    # INIT -------
    TARGET: type[NestInit_AnnotsAttrByKwArgs] | Any = AttrKit_Blank
    PATTS: tuple[str, ...] = ()

    def __init__(
            self,
            target: type | Any = None,
            patts: tuple[str, ...] = None,
            **kwargs,
    ) -> None | NoReturn:
        super().__init__(**kwargs)

        if target is not None:
            self.TARGET = target

        if patts is not None:
            self.PATTS = patts

    # -----------------------------------------------------------------------------------------------------------------
    def resolve(self) -> NestInit_AnnotsAttrByKwArgs | Any | NoReturn:
        # get dict -------
        data = dict(os.environ)     # just a copy!

        # filter ---
        if self.PATTS:
            filtered_out = filter(lambda name: not any([re.search(pat, name, flags=re.IGNORECASE) for pat in self.PATTS]), list(data))
            for out_i in filtered_out:
                data.pop(out_i)

        # load args -------
        if TypeAux(self.TARGET).check__class() and issubclass(self.TARGET, NestInit_AnnotsAttrByKwArgs):
            # used for check Annots all inited!

            result = self.TARGET(**data)
        else:
            AnnotsAux(self.TARGET).set_annots_attrs__by_args_kwargs(**data)
            result = self.TARGET

        return result



from typing import *

class PROJECT:
    # PROJECT ----------------------------------------------
    NAME_IMPORT: str = "testplans"
    KEYWORDS: list[str] = [
        "testplan",
        "testplan structure framework",
        "testplan gui",
        "testplan multy dut",
        "testplan several dut",
    ]

    # README -----------------------------------------------
    # add DOUBLE SPACE at the end of all lines! for correct representation in MD-viewers
    DESCRIPTION_SHORT: str = "simple testplan framework for several DUTs"
    DESCRIPTION_LONG: str = """
designed to apply testplan for several DUTs

## ПОНЯТИЯ
    TC - TestCase
    TP - TestPlan
    DUT - Device Under Test - тестируемое устройство

## АРХИТЕКТУРА
- тестплан
    - работает в потоке,
    - может быть остановлен в любой момент terminate(), при этом завершает все запущенные тесткейсы
    - имеет настройки которые принимаются всеми тесткейсами за базовые и могут быть перезаписаны ими для себя
    - имеет списки классов TC и обьектов DUT (генерирует обьекты TC для каждого DUT)
    - для себя не разделяет обьекты тесткейсов, работает строго с классом тесткейса,
    - выполняет все тесткейсы в порядке их следования на списке DUT
    - в один момент времени выполняется только один класс тесткейса
- тесткейсы
    - работают в потоке,
    - может быть остановлен в любой момент terminate(), при этом завершаются безопасно (исполняются все teardown обьектов и глобальный классовый тесткейса), 
    - представляет собой класс инициируемый с входным параметром DUT,
    - выполняются тесткейсы строго по очереди,
    - каждый тесткейс выполняется на всех устройствах либо асинхронно, либо синхронно в зависимости от настройки,
    - работа тесткейса полностью управляется классом тесткейса на серии устройств (возможно выполнение парных тестов с выбором нужных пар внутри тесткейса),
- результаты
    - все результаты находятся в пока в обьекте тесткейса
    - итогового (result)
    - промежуточных результатов (details)
- настройки
    - управление
        - SKIP всех возможных вариантов (полностью тесткейс для всех устройств, полностью DUT для всех TC, отдельный TC на отдельном DUT),
        - выполнение тесткейса синхронно/асинхронно
    - данные для использования в тесткейсах
        - реализовано в файлах JSON
        - для каждого тесткейса и общие для тестплана (кейсовые накладываются на плановые)
- GUI тестплана
    - запуск GUI опциональный,
    - старт/стоп тестплана,
    - отображение текущего тесткейса,
    - отображение результата тескейса на каждом тестируемом устройстве,
    - отображение промежуточных результатов (details)
- API 
    - минимальное API и запуск
"""

    FEATURES: list[str] = [
        # "feat1",
        # ["feat2", "block1", "block2"],

        ["[THREADS]",
            "safe work in independent TCs",
            "safe stop process at any moment by terminate",
         ],
        ["[SKIP]",
            "tc", "tc on dut", "dut"
         ],
        ["[DEVICES__BREEDER_INST]",
            "keep all in one instance",
            "use variants: single device for all duts or list for pairing each dut",
         ],
    ]

    # HISTORY -----------------------------------------------
    VERSION: tuple[int, int, int] = (0, 4, 34)
    TODO: list[str] = [
        "add meta for settings in tcs, it is better then applying in manually in TP!",
        "close all (api_server+tpThreads) on GUI close!",
        "add version for all jsons for future api_server",
        "[RESULTS] try separate",
    ]
    FIXME: list[str] = [
        "TP progress",
        "NEED TESTS!!! TC+TP"
    ]
    NEWS: list[str] = [
        "zero2",
    ]

# =====================================================================================================================



from typing import *
import uuid

from base_aux.testplans import *
from base_aux.breeders.m2_breeder_objects import *

from .tc import TestCaseBase
from .models import *


# =====================================================================================================================
class DeviceBase:
    # AUX -----------------------------------
    conn: Any = None
    INDEX: int = None

    # AUX -----------------------------------
    NAME: str = None
    DESCRIPTION: str = None
    SN: str = None
    DEV_FOUND: bool | None = None

    def __init__(self, index: int = None, **kwargs):
        """
        :param index: None is only for SINGLE!
        """
        if index is not None:
            self.INDEX = index
        super().__init__(**kwargs)

    # CONNECT ---------------------------------
    def connect(self) -> bool:
        if self.conn:
            try:
                return self.conn.connect()
            except:
                return False
        return True

    def disconnect(self) -> None:
        try:
            return self.conn.disconnect()
        except:
            pass

    def get__info__dev(self) -> dict[str, Any]:
        result = {
            "DUT_INDEX": self.INDEX,

            "DUT_NAME": self.NAME or self.__class__.__name__,
            "DUT_DESCRIPTION": self.DESCRIPTION or self.__class__.__name__,
            "DUT_SN": self.SN or "",
        }
        return result


# =====================================================================================================================
class DutBase(DeviceBase):
    SKIP: Optional[bool] = None

    def SKIP_reverse(self) -> None:
        """
        this is only for testing purpose
        """
        self.SKIP = not bool(self.SKIP)

    def _debug__reset_sn(self) -> None:
        """this is only for testing middleware"""
        self.SN = uuid.uuid4().hex


# =====================================================================================================================
class DevicesBreeder(BreederObjectList):
    def __del__(self):
        self.disconnect__cls()

    @classmethod
    def connect__cls(cls) -> None:
        cls.group_call__("connect")

    @classmethod
    def disconnect__cls(cls) -> None:
        cls.group_call__("disconnect")

    # DEBUG PURPOSE ---------------------------------------------------------------------------------------------------
    @classmethod
    def _debug__duts__reset_sn(cls) -> None:
        cls.group_call__("_debug__reset_sn", "DUT")


# =====================================================================================================================
class DevicesBreeder_WithDut(DevicesBreeder):
    """
    READY TO USE WITH DUT
    """
    # DEFINITIONS ---------------
    CLS_LIST__DUT: type[DutBase] = DutBase

    # JUST SHOW NAMES -----------
    LIST__DUT: list[DutBase]
    DUT: DutBase


# =====================================================================================================================
class DevicesBreeder_Example(DevicesBreeder_WithDut):
    """
    JUST an example DUT+some other single dev
    """
    # DEFINITIONS ---------------
    COUNT: int = 2
    CLS_SINGLE__ATC: type[DeviceBase] = DeviceBase

    # JUST SHOW NAMES -----------
    ATC: DeviceBase


# =====================================================================================================================



import datetime
import time

from base_aux.aux_attr.m1_attr1_aux import *
from base_aux.aux_cmp_eq.m1_cmp import *
from base_aux.aux_text.m1_text_aux import *
from base_aux.base_statics.m2_exceptions import *


# =====================================================================================================================
TYPE__TUPLE_DT_STYLE__DRAFT = tuple[str|None, str|None, str|None]
TYPE__TUPLE_DT_STYLE__FINAL = tuple[str, str, str]


class DateTimeStyle_Tuples:
    DT: TYPE__TUPLE_DT_STYLE__FINAL = ("-", " ", ":")       # default/standard from DateTime style for datetime.datetime.now()!
    DOTS: TYPE__TUPLE_DT_STYLE__FINAL = (".", " ", ".")     # same as DT but dots for data
    FILE: TYPE__TUPLE_DT_STYLE__FINAL = ("", "_", "")       # useful for filenames


# =====================================================================================================================
@final
class PatDateTimeFormat:
    def __init__(self, sep_date: str = None, sep_datetime: str = None, sep_time: str = None):
        """
        INIT separators only like schema
        """
        self.sep_date = sep_date or ""
        self.sep_datetime = sep_datetime or ""
        self.sep_time = sep_time or ""

    # -----------------------------------------------------------------------------------------------------------------
    @property
    def D(self) -> str:                                 # 2025-02-14 20250214 2025.02.14
        return f"%Y{self.sep_date}%m{self.sep_date}%d"

    @property
    def Dw(self) -> str:                                 # 2025-02-14-Mn 20250214Mn 2025.02.14.Mn
        """
        GOAL
        ----
        ensure weekDay
        """
        return f"%Y{self.sep_date}%m{self.sep_date}%d" + f"{self.sep_date}%a"

    # -----------------------------------------------------------------------------------------------------------------
    @property
    def T(self) -> str:                                 # 11:38:48
        return f"%H{self.sep_time}%M{self.sep_time}%S"

    @property
    def Tm(self) -> str:                                 # 11:38:48.442179
        """
        GOAL
        ----
        ensure ms
        """
        return f"%H{self.sep_time}%M{self.sep_time}%S" + ".%f"

    # -----------------------------------------------------------------------------------------------------------------
    @property
    def DT(self) -> str:
        return f"{self.D}{self.sep_datetime}{self.T}"      # 2025-02-14 11:38:48.442179

    @property
    def DTm(self) -> str:
        """
        GOAL
        ----
        ensure ms
        """
        return f"{self.D}{self.sep_datetime}{self.Tm}"      # 2025-02-14 11:38:48.442179

    @property
    def DwT(self) -> str:
        """
        GOAL
        ----
        ensure weekDay
        """
        return f"{self.Dw}{self.sep_datetime}{self.T}"      # 2025-02-14-Пн 11:38:48

    @property
    def DwTm(self) -> str:
        """
        GOAL
        ----
        ensure weekDay+ms
        """
        return f"{self.Dw}{self.sep_datetime}{self.Tm}"      # 2025-02-14-Пн 11:38:48.442179


# =====================================================================================================================
TYPE__DT_FINAL = datetime.datetime | datetime.date | datetime.time  # NOTE: dont use    | datetime.timedelta
TYPE__DT_DRAFT = TYPE__DT_FINAL | str | float | None    #  | int    # NOTE: int is not acceptable!


# =====================================================================================================================
# @final    # select styles
class DateTimeAux(NestCmp):
    SOURCE: TYPE__DT_FINAL = None
    STYLE: TYPE__TUPLE_DT_STYLE__FINAL = DateTimeStyle_Tuples.DOTS
    _PATTS: PatDateTimeFormat

    # patterns getattr -----
    D: str
    Dw: str
    T: str
    Tm: str

    DT: str
    DwT: str
    DTm: str
    DwTm: str

    # -----------------------------------------------------------------------------------------------------------------
    def __init__(self, source: TYPE__DT_DRAFT = None, style_tuple: TYPE__TUPLE_DT_STYLE__DRAFT = None) -> None | NoReturn:
        self.init_source(source)

        if style_tuple is not None:
            self.STYLE = style_tuple

        self._PATTS = PatDateTimeFormat(*self.STYLE)

    def init_source(self, source: TYPE__DT_DRAFT = None) -> None | NoReturn:
        if source is None:
            self.SOURCE = datetime.datetime.now()
        elif isinstance(source, (datetime.datetime, datetime.date, datetime.time, )):
            self.SOURCE = source
        elif isinstance(source, datetime.timedelta):
            # pass
            raise NotImplementedError(f"{source=}")
        elif isinstance(source, float):     # time.time()
            self.SOURCE = datetime.datetime.fromtimestamp(source)
        elif isinstance(source, int):
            raise NotImplementedError(f"{source=}")
        elif isinstance(source, str):
            self.SOURCE = self.parse_str(source, _raise=True)
        else:
            raise Exx__Incompatible(f"{source=}")

    @staticmethod
    def parse_str(source: str, _raise: bool = None) -> TYPE__DT_FINAL | None | NoReturn:
        nums = TextAux(source).findall(r"\d+")
        nums = list(map(int, nums))
        len_nums = len(nums)
        if len_nums in [6, 7]:
            result = datetime.datetime(*nums)
        elif len_nums == 4:
            result = datetime.time(*nums)
        elif len_nums == 3:
            if len(str(nums[0])) == 4:
                result = datetime.date(*nums)
            else:
                result = datetime.time(*nums)
        else:
            try:
                result = datetime.datetime.fromtimestamp(float(source))
            except:
                result = ""

        if not result and _raise:
            raise Exx__Incompatible(f"{source=}")
        return result

    # -----------------------------------------------------------------------------------------------------------------
    def __str__(self) -> str:
        return self.DT

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self})"

    def __int__(self):
        raise NotImplementedError()

    def __float__(self) -> float | NoReturn:
        return self.SOURCE.timestamp()

    # -----------------------------------------------------------------------------------------------------------------
    def __cmp__(self, other: Any) -> int | NoReturn:
        """
        GOAL
        ----
        appropriate CMP with time/date parts!
        """
        # todo: DEPRECATE??
        other = self.__class__(other)
        source1 = self.SOURCE
        source2 = other.SOURCE

        if isinstance(self.SOURCE, datetime.datetime) and isinstance(other.SOURCE, datetime.datetime):  # datetime FIRST!!!
            if source1 < source2:
                return -1
            elif source1 == source2:
                return 0
            elif source1 > source2:
                return 1

        elif isinstance(self.SOURCE, datetime.time) or isinstance(other.SOURCE, datetime.time):     # time SECOND!!!
            for attr in ["hour", "minute", "second", "microsecond"]:
                if getattr(source1, attr) < getattr(source2, attr):
                    return -1
                elif getattr(source1, attr) == getattr(source2, attr):
                    pass
                elif getattr(source1, attr) > getattr(source2, attr):
                    return 1
            return 0

        elif isinstance(self.SOURCE, datetime.date) or isinstance(other.SOURCE, datetime.date):     # date LAST!!! cause datetime(date)!
            for attr in ["year", "month", "day"]:
                if getattr(source1, attr) < getattr(source2, attr):
                    return -1
                elif getattr(source1, attr) == getattr(source2, attr):
                    pass
                elif getattr(source1, attr) > getattr(source2, attr):
                    return 1
            return 0

        else:
            raise NotImplementedError()

    # -----------------------------------------------------------------------------------------------------------------
    def get_str__by_pat(self, pattern: str) -> str:
        """
        NOTE
        ----
        mainly used internal!

        GOAL
        ----
        make str by pat

        EXAMPLES
        --------
        %Y%m%d_%H%M%S -> 20241203_114845
        add_ms -> 20241203_114934.805854
        """
        return self.SOURCE.strftime(pattern)

    def __getattr__(self, item: str) -> str | NoReturn:
        if item in AttrAux(PatDateTimeFormat).iter__not_hidden():
            if isinstance(self.SOURCE, datetime.datetime):
                pass
            elif isinstance(self.SOURCE, datetime.date):
                item = TextAux(item).clear__regexps("T", "m", flags=re.IGNORECASE)
            elif isinstance(self.SOURCE, datetime.time):
                item = TextAux(item).clear__regexps("D", "w", flags=re.IGNORECASE)

            return self.get_str__by_pat(pattern=getattr(self._PATTS, item))
        else:
            raise AttributeError(item)


# =====================================================================================================================
@final
class DateTimeAuxDT(DateTimeAux):
    STYLE: TYPE__TUPLE_DT_STYLE__FINAL = DateTimeStyle_Tuples.DT


@final
class DateTimeAuxDOTS(DateTimeAux):
    STYLE: TYPE__TUPLE_DT_STYLE__FINAL = DateTimeStyle_Tuples.DOTS


@final
class DateTimeAuxFILE(DateTimeAux):
    STYLE: TYPE__TUPLE_DT_STYLE__FINAL = DateTimeStyle_Tuples.FILE


# =====================================================================================================================
if __name__ == '__main__':
    print(repr(DateTimeAux()))
    print(str(DateTimeAux()))
    print()

    print(DateTimeAux().T)
    print(DateTimeAux().Tm)
    print()

    print(DateTimeAux().D)
    print(DateTimeAux().DT)
    print(DateTimeAux().DwTm)
    print()

    inst = DateTimeAux(datetime.date(2024, 2, 1))
    print(inst)
    print(inst.DT)
    print(inst.DwTm)

    inst = DateTimeAux(datetime.time(11, 50, 1, 123))
    print(inst)
    print(inst.DT)
    print(inst.DwTm)

    # inst = DateTimeAux(datetime.timedelta(11, 50, 1, 123))
    # print(inst)
    # print(inst.DT)
    # print(inst.DwTm)

    inst = DateTimeAux(datetime.datetime.now().date())
    print(inst)
    print(inst.DT)
    print(inst.DwTm)

    print(time.time())  # float!

    # print(datetime.date(2024, 2, 1).timestamp())
    print(DateTimeAux("2025.02.26 17.00.56"))


# =====================================================================================================================



import time
from typing import *


# =====================================================================================================================
@final
class Sleep:
    """
    GOAL
    ----
    just a primitive func for tests or other purpose!
    """
    SEC: float = 1

    def __init__(self, sec: float = None):
        if sec is not None:
            self.SEC = sec

    def echo(self, echo: Any = None, *args, **kwargs) -> Any:
        time.sleep(self.SEC)
        return echo

    def NONE(self, *args, **kwargs) -> None:
        return self.echo(echo=None, *args, **kwargs)

    def TRUE(self, *args, **kwargs) -> bool:
        return self.echo(echo=True, *args, **kwargs)

    def FALSE(self, *args, **kwargs) -> bool:
        return self.echo(echo=False, *args, **kwargs)

    def EXX(self, *args, **kwargs) -> Exception:
        return self.echo(echo=Exception("Sleep.EXX"), *args, **kwargs)

    def RAISE(self, *args, **kwargs) -> NoReturn:
        time.sleep(self.SEC)
        raise Exception("Sleep.RAISE")


# =====================================================================================================================
